#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include "ippc/ippc.h"
#include "ippc/syscall.h"
#include "api/horde.h"

void _session_init();
void _stream_init(int handle, horde_boolean connected);

//
// debug logging stuff
//
static int current_log_level = 1;

#define HORDE_LOG(LEVEL,...)\
  if (LEVEL <= current_log_level) {\
    int k;\
    for (k = 0; k < LEVEL; k++) fprintf(stderr,"  ");\
    fprintf(stderr,__VA_ARGS__); } else { }


void horde_set_log_level(int level) {
  current_log_level = level;

  HORDE_LOG(LOG_LEVEL_DEBUG, "changed log level to %d.\n", level);
}


//
// IPPC stuff
//
ippc_channel ippcc = INVALID_CHANNEL;
const char *ippc_namespace_root = "/var/horde";

void horde_set_ippc_namespace_root(const char* root) {
  ippc_namespace_root = root;
}

ippc_channel horde_ippc_init() {
  char pathstr[256];
  snprintf(pathstr, 256, "%s/sessions/ippc", ippc_namespace_root);

  if (ippcc) ippc_channel_close(ippcc, FALSE);
  ippcc = ippc_channel_open(pathstr);

  return ippcc;
}

int horde_get_ippc_fd() {
  if (ippcc != INVALID_CHANNEL)
    return ippc_channel_get_fd(ippcc);
  else 
    return -1;
}

ippc_channel 
horde_session_ippc_init(const char* session_id) {
  char pathstr[256];
  snprintf(pathstr, 256, "%s/sessions/%s/ippc", ippc_namespace_root, session_id);

  if (ippcc) ippc_channel_close(ippcc, FALSE);
  ippcc = ippc_channel_open(pathstr);

  return ippcc;
}

//
// session stuff
//
horde_boolean
_startup_session(const char* caller, ippcmsg_session_command* com) {
  int type;
  size_t len;  
  char *reply = ippc_call(ippcc, 
			  IPPC_SESSION_COMMAND, com, sizeof(struct ippcmsg_session_command),
			  IPPC_SESSION_EVENT, &type, &len);

  if (reply) {
    if (type == IPPC_SESSION_EVENT) {
      ippcmsg_session_event* evt = (ippcmsg_session_event*)reply;

      if (evt->event == evt->CONNECTED) {
	_session_init(); // initialize event handler stuff
	return TRUE;
      } else { 
	HORDE_LOG(LOG_LEVEL_ERROR, 
		  "%s() received unexpected session event: %d\n", caller, (int)evt->event);
	return FALSE;
      }
    } else { // exception
      ippcmsg_exception* ex = (ippcmsg_exception*)reply;

      HORDE_LOG(LOG_LEVEL_ERROR, 
		"IPPC exception in %s(): %s [%d]\n", caller, ex->msg, ex->code);

      return FALSE;
    }
  } else {
    HORDE_LOG(LOG_LEVEL_ERROR, 
	      "ippc_call() failed in %s(): %s\n", caller, strerror(errno));
    return FALSE;
  }  
}

horde_boolean 
horde_session_join(const char* session_name) {
  // init IPPC
  if (horde_session_ippc_init(session_name) == INVALID_CHANNEL) {
    HORDE_LOG(LOG_LEVEL_ERROR, 
	      "session_join() couldn't init IPPC: %s, ippc path: %s/sessions/%s/ippc\n", 
	      strerror(errno), ippc_namespace_root, session_name);
    return FALSE;
  }

  // join
  ippcmsg_session_command com;
  bzero((char*)&com, sizeof (struct ippcmsg_session_command));
  com.command = com.JOIN;
  strncpy(com.session_id, session_name, SESSION_ID_MAX_LENGTH);

  return _startup_session("session_join", &com);
}

horde_boolean 
horde_session_connect(const char* connect_address, const char* session_name) {
  // init IPPC
  if (horde_ippc_init() == INVALID_CHANNEL) {
    HORDE_LOG(LOG_LEVEL_ERROR, 
	      "session_connect() couldn't init IPPC: %s, ippc path: %s/sessions/ippc\n", 
	      strerror(errno), ippc_namespace_root);
    return FALSE;
  }

  // connect
  ippcmsg_session_command com;
  bzero((char*)&com, sizeof (struct ippcmsg_session_command));
  com.command = com.CONNECT;
  strncpy(com.session_id, session_name, SESSION_ID_MAX_LENGTH);
  strncpy(com.peer_address, connect_address, NET_ADDRESS_MAX_LENGTH);

  return _startup_session("session_connect", &com);
}

horde_boolean 
horde_session_accept(const char* session_name) {
  // init IPPC
  if (horde_ippc_init() == INVALID_CHANNEL) {
    HORDE_LOG(LOG_LEVEL_ERROR, 
	      "session_accept() couldn't init IPPC: %s, ippc path: %s/sessions/ippc\n", 
	      strerror(errno), ippc_namespace_root);
    return FALSE;
  }

  // accept
  ippcmsg_session_command com;
  bzero((char*)&com, sizeof (struct ippcmsg_session_command));
  com.command = com.ACCEPT;
  strncpy(com.session_id, session_name, SESSION_ID_MAX_LENGTH);

  return _startup_session("session_accept", &com);
}

void
horde_session_close(const char* session_name) {
  if (ippcc != INVALID_CHANNEL) {
      ippcmsg_session_command com;
      bzero((char*)&com, sizeof (struct ippcmsg_session_command));
      com.command = com.CLOSE;
      strncpy(com.session_id, session_name, SESSION_ID_MAX_LENGTH);
      
      ippc_message_send(ippcc, IPPC_SESSION_COMMAND, &com, sizeof(struct ippcmsg_session_command));

      // let the session event handler deal with the reply...
  }
}

//
// stream connect/accept/close
//

void
horde_stream_close(const char* stream_id, int handle) {
  if (ippcc != INVALID_CHANNEL) {
      ippcmsg_stream_command com;
      bzero((char*)&com, sizeof (struct ippcmsg_stream_command));
      com.command = com.CLOSE;
      strncpy(com.stream_id, stream_id, STREAM_ID_MAX_LENGTH);
      com.stream_uid = handle;
      
      ippc_message_send(ippcc, IPPC_STREAM_COMMAND, &com, sizeof(struct ippcmsg_stream_command));      
      // let the session event handler deal with the exception if it occurs...

      _stream_init(handle, FALSE); // clear event handler state
  }
}

int
_startup_stream(const char* caller, ippcmsg_stream_command* com) {
  int type;
  size_t len;  
  char *reply = ippc_call(ippcc, 
			  IPPC_STREAM_COMMAND, com, sizeof(struct ippcmsg_stream_command),
			  IPPC_STREAM_EVENT, &type, &len);

  if (reply) {
    if (type == IPPC_STREAM_EVENT) {
      ippcmsg_stream_event* evt = (ippcmsg_stream_event*)reply;

      if (evt->event == evt->CONNECTED) {
	 _stream_init(evt->stream_uid, TRUE); // init event handler state
	return evt->stream_uid;
      } else { 
	HORDE_LOG(LOG_LEVEL_ERROR, 
		  "%s() received unexpected stream event: %d\n", caller, (int)evt->event);
	return -1;
      }
    } else { // exception
      ippcmsg_exception* ex = (ippcmsg_exception*)reply;

      HORDE_LOG(LOG_LEVEL_ERROR, 
		"IPPC exception in %s(): %s [%d]\n", caller, ex->msg, ex->code);

      return -1;
    }
  } else {
    HORDE_LOG(LOG_LEVEL_ERROR, 
	      "ippc_call() failed in %s(): %s\n", caller, strerror(errno));
    return -1;
  }  
}

int
horde_stream_accept(const char* stream_id) {
  if (ippcc != INVALID_CHANNEL) {

    // accept
    ippcmsg_stream_command com;
    bzero((char*)&com, sizeof (struct ippcmsg_stream_command));
    com.command = com.ACCEPT;
    strncpy(com.stream_id, stream_id, STREAM_ID_MAX_LENGTH);  
    
    return _startup_stream("stream_accept", &com);
  }
}

int
horde_stream_connect(const char* stream_id) {
  if (ippcc != INVALID_CHANNEL) {

    // connect
    ippcmsg_stream_command com;
    bzero((char*)&com, sizeof (struct ippcmsg_stream_command));
    com.command = com.CONNECT;
    strncpy(com.stream_id, stream_id, STREAM_ID_MAX_LENGTH);
    
    return _startup_stream("stream_connect", &com);
  }
}

int horde_stream_connect_multiple(int nstreams, const char **id, int *handles) {
  int conn = 0;
  if (ippcc != INVALID_CHANNEL) {
    // send n connect messages
    ippcmsg_stream_command com;
    bzero((char*)&com, sizeof (struct ippcmsg_stream_command));
    for (int i = 0; i < nstreams; i++) {
      com.command = com.CONNECT;
      strncpy(com.stream_id, id[i], STREAM_ID_MAX_LENGTH);

      ippc_message_send(ippcc, IPPC_STREAM_COMMAND, &com, sizeof(struct ippcmsg_stream_command));
    }

    // init stream handles to invalid states
    for (int i = 0; i < nstreams; i++) {
      handles[i] = -1;
    }

    // read n messages
    char *msg;
    int type; 
    size_t length;
    for (int i = 0; i < nstreams; i++) {
      msg = ippc_message_receive(ippcc, &type, &length);

      if (type == IPPC_STREAM_EVENT) {
	ippcmsg_stream_event* evt = (ippcmsg_stream_event*)msg;
	
	if (evt->event == evt->CONNECTED) {
	  for (int j = 0; j < nstreams; j++) {
	    if (strcmp(id[j], evt->stream_id) == 0) handles[j] = evt->stream_uid;
	  }

	  _stream_init(evt->stream_uid, TRUE); // init event handler state
	  conn++;
	} else { 
	  HORDE_LOG(LOG_LEVEL_ERROR, 
		    "ignoring unexpected stream event: %d\n", (int)evt->event);	  
	}
      } else { // exception
	ippcmsg_exception* ex = (ippcmsg_exception*)msg;
	
	HORDE_LOG(LOG_LEVEL_ERROR, 
		  "IPPC exception in stream_connect_multiple(): %s [%d]\n", ex->msg, ex->code);
	conn++;
      }
    }
  }
  return conn;
}

int horde_stream_accept_multiple(int nstreams, const char **id, int *handles) {
  int conn = 0;
  if (ippcc != INVALID_CHANNEL) {
    // send n accept messages
    ippcmsg_stream_command com;
    bzero((char*)&com, sizeof (struct ippcmsg_stream_command));
    for (int i = 0; i < nstreams; i++) {
      com.command = com.ACCEPT;
      strncpy(com.stream_id, id[i], STREAM_ID_MAX_LENGTH);

      ippc_message_send(ippcc, IPPC_STREAM_COMMAND, &com, sizeof(struct ippcmsg_stream_command));
    }

    // init stream handles to invalid states
    for (int i = 0; i < nstreams; i++) {
      handles[i] = -1;
    }

    // read n messages
    char *msg;
    int type; 
    size_t length;
    for (int i = 0; i < nstreams; i++) {
      msg = ippc_message_receive(ippcc, &type, &length);

      if (type == IPPC_STREAM_EVENT) {
	ippcmsg_stream_event* evt = (ippcmsg_stream_event*)msg;
	
	if (evt->event == evt->CONNECTED) {
	  for (int j = 0; j < nstreams; j++) {
	    if (strcmp(id[j], evt->stream_id) == 0) handles[j] = evt->stream_uid;
	  }

	  _stream_init(evt->stream_uid, TRUE); // init event handler state
	  conn++;
	} else { 
	  HORDE_LOG(LOG_LEVEL_ERROR, 
		    "ignoring unexpected stream event: %d\n", (int)evt->event);	  
	}
      } else { // exception
	ippcmsg_exception* ex = (ippcmsg_exception*)msg;
	
	HORDE_LOG(LOG_LEVEL_ERROR, 
		  "IPPC exception in stream_accept_multiple(): %s [%d]\n", ex->msg, ex->code);
	conn++;
      }
    }
  }
  return conn;
}

void
horde_stream_set_options(int stream_handle, int opts) {
  if (ippcc != INVALID_CHANNEL) {
    ippcmsg_stream_command com;
    bzero((char*)&com, sizeof (struct ippcmsg_stream_command));
    com.command = com.SET_EVENT_MASK;
    com.stream_uid = stream_handle;
    
    com.event_mask_ack = opts & STREAM_OPT_EVENT_ACK;
    com.event_mask_loss = opts & STREAM_OPT_EVENT_LOSS;
    com.event_mask_recv = opts & STREAM_OPT_EVENT_RECV;
    com.event_mask_choke = opts & STREAM_OPT_IO_PRESSURE;
    com.event_mask_flushed = opts & STREAM_OPT_EVENT_FLUSHED;

    ippc_message_send(ippcc, IPPC_STREAM_COMMAND, &com, sizeof(struct ippcmsg_stream_command));
  }
}

void horde_stream_send(int hstream, int adu_id, const char* adu_data, size_t data_sz, horde_objective_tags tags) {
  size_t n = data_sz + sizeof(struct ippcmsg_data_send);
  ippcmsg_data_send* msg = (ippcmsg_data_send*)calloc(1,n);

  msg->hstream = hstream;
  msg->adu_id = adu_id;
  msg->data_length = data_sz;

  msg->is_fragment = false;
  msg->adu_offset = 0;
  msg->adu_length = data_sz;

  msg->tags = tags;

  memcpy(msg->data, adu_data, data_sz);

  ippc_message_send(ippcc, IPPC_DATA_SEND, msg, n);

  free(msg);
}

void horde_stream_send_fragment(int hstream, int adu_id, int adu_offset, size_t adu_length,
				const char* fragment_data, size_t fragment_sz, horde_objective_tags tags) {
  size_t n = fragment_sz + sizeof(struct ippcmsg_data_send);
  ippcmsg_data_send* msg = (ippcmsg_data_send*)calloc(1,n);

  msg->hstream = hstream;
  msg->adu_id = adu_id;
  msg->data_length = fragment_sz;

  msg->is_fragment = true;
  msg->adu_offset = adu_offset;
  msg->adu_length = adu_length;

  msg->tags = tags;

  memcpy(msg->data, fragment_data, fragment_sz);

  ippc_message_send(ippcc, IPPC_DATA_SEND, msg, n);

  free(msg);
}


void
horde_stream_clear(int stream_handle) {
  if (ippcc != INVALID_CHANNEL) {
    ippcmsg_stream_command com;
    bzero((char*)&com, sizeof (struct ippcmsg_stream_command));
    com.command = com.CLEAR;
    com.stream_uid = stream_handle;
    
    ippc_message_send(ippcc, IPPC_STREAM_COMMAND, &com, sizeof(struct ippcmsg_stream_command));
  }
}

void
horde_statistics_enable(horde_boolean enable) {
  if (ippcc != INVALID_CHANNEL) {
    ippcmsg_session_command com;
    bzero((char*)&com, sizeof (struct ippcmsg_session_command));
    com.command = (enable) ? com.ENABLE_STATISTICS : com.DISABLE_STATISTICS;
    
    ippc_message_send(ippcc, IPPC_SESSION_COMMAND, &com, sizeof(struct ippcmsg_session_command));
  }
}

//
// event handler stuff
//
horde_callback_session_t session_cb = NULL;
horde_callback_exception_t exception_cb = NULL;
horde_callback_statistics_t statistics_cb = NULL;

horde_boolean stream_connected[256];

horde_callback_recv_t recv_callback[256];
void* recv_cb_context[256];

horde_callback_stream_t stream_callback[256];
void* stream_cb_context[256];

void default_exception_cb(horde_exception_code code, const char* msg) {
  printf("Exception %d in horde API: '%s'\n", (int)code, msg);
}

horde_boolean is_stream_connected(int handle) {
  if ((handle > 0) && (handle < 256))
    return stream_connected[handle];
  else
    return FALSE;
}

void horde_session_set_callback(horde_callback_session_t cb) {
  session_cb = cb;
}

void horde_exception_set_callback(horde_callback_exception_t cb) {
  exception_cb = cb;
}

void horde_statistics_set_callback(horde_callback_statistics_t cb) {
  statistics_cb = cb;
}

void horde_stream_set_recv_callback(int hstream, horde_callback_recv_t cb, void* cb_context) { 
  if (is_stream_connected(hstream)) {
    recv_callback[hstream] = cb;
    recv_cb_context[hstream] = cb_context;
  }
}

void* horde_stream_get_recv_cb_context(int hstream) {
  if (is_stream_connected(hstream)) {
    return recv_cb_context[hstream];
  } else {
    return NULL;
  }
}

void horde_stream_set_event_callback(int hstream, horde_callback_stream_t cb, void* cb_context) {
  if (is_stream_connected(hstream)) {
    stream_callback[hstream] = cb;
    stream_cb_context[hstream] = cb_context;
  }
}

void* horde_stream_get_event_cb_context(int hstream) {
  if (is_stream_connected(hstream)) {
    return stream_cb_context[hstream];
  } else {
    return NULL;
  }
}

void _session_init() {
  session_cb = NULL;
  exception_cb = default_exception_cb;
  statistics_cb = NULL;

  memset(stream_connected, 0, 256 * sizeof(horde_boolean));
  memset(recv_callback, 0, 256 * sizeof(horde_callback_recv_t));
  memset(recv_cb_context, 0, 256 * sizeof(void*));
  memset(stream_callback, 0, 256 * sizeof(horde_callback_stream_t));
  memset(stream_cb_context, 0, 256 * sizeof(void*));
}

void _stream_init(int handle, horde_boolean connected) {
  if ((handle >= 0) && (handle < 256)) {
    stream_connected[handle] = connected; 
    recv_callback[handle] = NULL;
    recv_cb_context[handle] = NULL;
    stream_callback[handle] = NULL;
    stream_cb_context[handle] = NULL; 
  }
}

void dispatch_session_event(char *msg) {
  ippcmsg_session_event* evt = (ippcmsg_session_event*)msg;
  if (session_cb != NULL) {
    if (evt->event == (*evt).CLOSED) {
      session_cb(SESSION_CLOSED);
    }
  }  
}

void dispatch_stream_event(char *msg) {
  ippcmsg_stream_event* evt = (ippcmsg_stream_event*)msg;
  if (is_stream_connected(evt->stream_uid) && (stream_callback[evt->stream_uid] != NULL)) {
    if (evt->event == (*evt).CLOSED) { 
      stream_callback[evt->stream_uid](STREAM_CLOSED, evt->stream_uid, 0, 0, 0, 0, 
				       stream_cb_context[evt->stream_uid]);
    } else if (evt->event == (*evt).FLUSHED) {
      stream_callback[evt->stream_uid](STREAM_FLUSHED, evt->stream_uid, 0, 0, 0, 0,
				       stream_cb_context[evt->stream_uid]);
    }
  }
}

void dispatch_datafb_event(char *msg) {
  ippcmsg_data_feedback* fb = (ippcmsg_data_feedback*)msg;
  if (is_stream_connected(fb->stream_uid) && (stream_callback[fb->stream_uid] != NULL)) {
    horde_stream_event evt = (fb->type == (*fb).ACK) ? ADU_FRAG_ACK : ADU_FRAG_LOSS;
    stream_callback[fb->stream_uid](evt, fb->stream_uid, fb->adu_id, fb->frag_start, fb->frag_length, 
				    fb->rtt_ms, stream_cb_context[fb->stream_uid]);  
  }
}

void dispatch_datarecv_event(char *msg) {
  ippcmsg_data_recv* recv = (ippcmsg_data_recv*)msg;
  if (is_stream_connected(recv->hstream) && (recv_callback[recv->hstream] != NULL)) {
    recv_callback[recv->hstream](recv->hstream, recv->adu_id, recv->total_length, 
				 recv->frag_start, recv->frag_length, recv->data, 
				 recv_cb_context[recv->hstream]);
  }  
}

void dispatch_ex_event(char *msg) {
  ippcmsg_exception* ex = (ippcmsg_exception*)msg;
  if (exception_cb) {     
    exception_cb((horde_exception_code)ex->code, ex->msg);
  } else {
    HORDE_LOG(LOG_LEVEL_DEBUG, "dispatch_ex_event: no callback, ignoring exception '%s' [%d]\n", ex->msg, (int)ex->code);
  }
}

void dispatch_stat_event(char *msg) {
  if (statistics_cb) { 
    statistics_report* report = (statistics_report*)msg; ///@todo fix this hack...
    statistics_cb(report);
  } else {
    HORDE_LOG(LOG_LEVEL_DEBUG, "dispatch_stat_event: no registered callback.\n");
  }
}

int horde_event_dispatch(int timeout_ms) {
  if (ippcc != INVALID_CHANNEL) {
    // try to receive a message
    char *msg;
    int type; 
    size_t length;
    
    if (timeout_ms < 0) {
      msg = ippc_message_receive(ippcc, &type, &length);
    } else {
      msg = ippc_message_receive_timed(ippcc, &type, &length, timeout_ms);
    }

    if ((errno != 0) && (errno != EAGAIN)) {
      HORDE_LOG(LOG_LEVEL_DEBUG, "event_dispatch: unexpected error: '%s' [%d]\n", strerror(errno), errno);
      return -1;
    } else if ((msg == NULL) || (msg == IPPC_RECV_INCOMPLETE) || (errno == EAGAIN)) return 0;
    else {
      // got a message...
      HORDE_LOG(LOG_LEVEL_DEBUG, "event_dispatch: got message, type=%d\n", type);

      switch (type) {
      case IPPC_SESSION_EVENT:
	dispatch_session_event(msg);
	break;
      case IPPC_STREAM_EVENT:
	dispatch_stream_event(msg);
	break;
      case IPPC_DATA_FEEDBACK:
	dispatch_datafb_event(msg);
	break;
      case IPPC_DATA_RECV:
	dispatch_datarecv_event(msg);
	break;
      case IPPC_EXCEPTION:
	dispatch_ex_event(msg);
	break;
      case IPPC_STATISTICS:
	dispatch_stat_event(msg);
      default:
	// ignore all other messages...
	break;
      }
      return 1;
    }
  } else {
    return -1;
  }
}

