/** @file horde.h
 *
 * C-language interface to the horde daemon.
 *
 * This API doesn't use event callbacks or multiple
 * threads for simplicity.
 *
 */
#ifndef __HORDE_C_API_H__
#define __HORDE_C_API_H__

#include "shared/objectives.h"

/** A useful enum: boolean, independent of whether we're in C or C++ */
typedef enum { FALSE=0, TRUE=1 } horde_boolean;

#define LOG_LEVEL_DEBUG 3
#define LOG_LEVEL_INFO 2
#define LOG_LEVEL_ERROR 1
#define LOG_LEVEL_NONE 0

/** 
 * Set the API's debug log's verbosity level. Messages are logged to
 * \a stderr. Setting the level to \a LOG_LEVEL_NONE turns off
 * logging; \a LOG_LEVEL_ERROR only displays error messages (this is
 * the default); \a LOG_LEVEL_DEBUG turns on everything.
 */
void horde_set_log_level(int level);

/** 
 * Set the root name for the IPPC namespace (eg \c "/var/horde").  A
 * pointer to \a root may be saved, so don't deallocate or reuse that
 * memory.
 */
void horde_set_ippc_namespace_root(const char* root);

/**
 * Get the file descriptor for the socket being used to communicate with 
 * the horde IPPC daemon. Nothing should be written to or read from this
 * descriptor. This function is provided so that the descriptor can
 * be used inside an external select loop. 
 *
 * @return -1 or a valid file descriptor.
 */
int horde_get_ippc_fd();

//
// SESSION
//

/**
 * Wait for and accept a session request from a remote host, if the
 * name presented by the remote host matches this name. This method
 * blocks until a connection request is received and resolved, or an
 * error occurs (e.g., connection broken with horde backend).
 *
 * @param  session_name the session name (null-terminated string)
 * @return TRUE on success, or FALSE if an error occurs and the session could
 *         not be estalished.
 */
horde_boolean horde_session_accept (const char* session_name);

/**
 * Attempts to connect to a remote host, using the given session
 * name. The remote host must be waiting to accept a session. This
 * blocks until it is able to connect, or an error occurs.
 *
 * @param connect_address the address of the remote host to connect to
 *          (can be a hostname, or an IP address expressed in dotted notation).
 * @param  session_name the name of the session (null-terminated string)
 * @return TRUE on success, or FALSE if an error occurs and the session could
 *         not be estalished.
 */
horde_boolean horde_session_connect(const char* connect_address, const char* session_name);

/**
 * Closes the named session, if it is open, or does nothing.
 */
void horde_session_close(const char* session_name);

/**
 * Joins an already connected session.
 *
 * @param  session_name the session name (null-terminated string)
 * @return TRUE on success, or FALSE if an error occurs and the session could
 *         not be joined.
 */
horde_boolean horde_session_join(const char* session_name);

//
// STREAM
//

/**
 * Waits for and handles a stream connect request. Blocks until a
 * stream with the given identifier is accepted or an error
 * occurs. Returns a handle for the open stream, or -1,
 * and sets errcode.
 *
 * @param stream_id case sensitive stream identifier (NULL terminated).
 * @return a valid stream handle (>0) or -1.
 */
int horde_stream_accept(const char* stream_id);

/**
 * Sends a stream connect request. Blocks until the stream is
 * connected or an error occurs. Returns a handle to the open stream,
 * or -1.
 *
 * @param stream_id case sensitive stream identifier (NULL terminated).
 * @return a valid stream handle (>0) or -1.
 */
int horde_stream_connect(const char* stream_id);


/**
 * Roughly equivalent to call horde_stream_accept N-times, but potentially far more
 * efficent.
 *
 * @param nstreams the number of entries in id and handles
 * @param id an array of string identifiers (all unique)
 * @param handles an array which will be filled in with the stream handle for the corresponding id.
 * @returns the number of streams succesfully accepted. all streams that couldn't connect,
 *    will have a handle equal to -1 when this function returns.
 */
int horde_stream_accept_multiple(int nstreams, const char **id, int *handles);

/**
 * Roughly equivalent to call horde_stream_connect N-times, but potentially far more
 * efficent.
 *
 * @param nstreams the number of entries in id and handles
 * @param id an array of string identifiers (all unique)
 * @param handles an array which will be filled in with the stream handle for the corresponding id.
 * @returns the number of streams succesfully accepted. all streams that couldn't connect,
 *    will have a handle equal to -1 when this function returns.
 */
int horde_stream_connect_multiple(int nstreams, const char **id, int *handles);

/**
 * Closes an open stream, given the stream's case sensitive identifier
 * and a handle to it.
 *
 * @param stream_id case sensitive stream identifier (NULL terminated).
 * @param handle a handle returned by \a horde_stream_accept() or \a horde_stream_open().
 */ 
void horde_stream_close(const char* stream_id, int handle);

static const int STREAM_OPT_NONE           = 0x00000000;
static const int STREAM_OPT_EVENT_RECV     = 0x00000001;
static const int STREAM_OPT_EVENT_ACK      = 0x00000002;
static const int STREAM_OPT_EVENT_LOSS     = 0x00000004;
static const int STREAM_OPT_EVENT_FLUSHED  = 0x00000008;
static const int STREAM_OPT_IO_PRESSURE    = 0x00000010;
static const int STREAM_OPT_ALL            = 0xffffffff;

/**
 * Sets stream options. When a stream is created, by default all
 * options are set.
 *
 * Most of the stream options are event masks (when they are not set,
 * that type of event is not delivered to the stream event
 * handler). \a STREAM_OPT_IO_PRESSURE is the only complex option. When
 * this option is set (default), sending too much data will cause the
 * IPPC socket to block, to not accept any new IPPC sends, until the
 * data already queued inside the horde daemon has been sent on the
 * network. When \a STREAM_OPT_IO_PRESSURE is not set, the IPPC layer will
 * accept data as fast as it can. The horde daemon may silently drop
 * packets internally if the client is sending at a rate much higher than
 * that allocated (see the throttle event).
 *
 * @param opts A set of option flags. A set of STREAM_OPT_ values OR'd
 * together. Any STREAM_OPT_ that has not been added to opts, will be
 * disabled, and options set in opts will be enabled. If you want to
 * turn off a single option, start with STREAM_OPT_ALL and XOR out any
 * options you don't want.
 */
void horde_stream_set_options(int stream_handle, int opts=STREAM_OPT_ALL);

/**
 * Send an ADU.
 *
 * May block. See the \a STREAM_OPT_IO_PRESSURE notes for
 * horde_stream_set_options().
 *
 * @note Be careful: if the ADU is larger than the maximum packet
 * size, and fragmentation is disallowed, the stream may get blocked
 * up, and no data after the large ADU may ever be sent. Depends
 * on what scheduler is being used.
 */
void horde_stream_send(int hstream, int adu_id, const char* adu_data, size_t data_sz,
		       horde_objective_tags tags);

/**
 * Send an ADU fragment. Useful for retransmissions.
 */
void horde_stream_send_fragment(int hstream, int adu_id, int adu_offset, size_t adu_length,
				const char* fragment_data, size_t fragment_sz, horde_objective_tags tags);

/**
 * Clears a stream's send queue.
 *
 * If any ADU's are waiting in the send queue, they are silently discarded.
 */
void horde_stream_clear(int hstream);


//
// event callback stuff
//

enum horde_session_event {
SESSION_CLOSED=1 
};

typedef void (*horde_callback_session_t)(enum horde_session_event);

typedef void (*horde_callback_recv_t)(int, int, int, int, int, const char*, void*);
// callback(stream_uid, adu_id, total_len, frag_offset, frag_length, frag_data, ptr_context)

enum horde_stream_event {
  UNUSED=0,
  ADU_FRAG_ACK=1,
  ADU_FRAG_LOSS=2,
  STREAM_CLOSED=3,
  STREAM_FLUSHED=4
};

typedef void (*horde_callback_stream_t)(enum horde_stream_event, int, int, int, int, int, void*);
// callback(event_code, stream_uid, a, b, c, d, ptr_context)
// a=adu_id (for ACK, LOSS events)
// b=fragment_start (for ACK/LOSS events)
// c=fragment_length (for ACK/LOSS events)
// d=RTT (for ACK events)
// ptr_context=cb_context from horde_stream_set_event_callback

enum horde_exception_code {
  EX_BAD_IPPC=1,
  EX_ILLEGAL_OPERATION=2,
  EX_INSUFFICIENT_PERMISSION=3,
  EX_BAD_SESSION_ID=4,
  EX_BAD_STREAM_ID=5,
  EX_INTERNAL=255
};

typedef void (*horde_callback_exception_t)(enum horde_exception_code, const char*);

/**
 * @param cb the callback to be invoked on session events.
 */
void horde_session_set_callback(horde_callback_session_t cb);

void horde_exception_set_callback(horde_callback_exception_t cb);

/**
 * Set the callback to be invoked on stream data recv events
 */
void horde_stream_set_recv_callback(int hstream, 
				    horde_callback_recv_t cb, void* cb_context);

/**
 * Set the callback to be invoked on stream control events.
 */
void horde_stream_set_event_callback(int hstream, 
				     horde_callback_stream_t cb, void* cb_context);

// for python bindings...
void* horde_stream_get_recv_cb_context(int hstream);
void* horde_stream_get_event_cb_context(int hstream);

/**
 * Blocks, waits for an event for some maximum amount of time, and if
 * an event occurs during that period, it calls the appropriate
 * callback, and returns immediately.
 *
 * @param timeout_ms maximum amount of time to block waiting for an event.
 *                   <0 implies an infinite timeout (default behaviour)
 * @return <0 on error, 0 if no events, or 1 if an event was dispatched
 */
int horde_event_dispatch(int timeout_ms=-1);

// utility.
//const char* to_hexstr(const char* data, size_t datalen);

//
// statistics stuff
//

/**
 * Enables or disables statistics reporting. If enabled, the statistics
 * event handler will be called periodically with network statistics.
 */
void horde_statistics_enable(horde_boolean enable=TRUE);

#ifndef __HORDE_IPPC_SYSCALL_H__
#define CHANNEL_ID_MAX_LENGTH 31
#define STAT_MAX_CHANNELS 16
#endif

struct statistics_channel { // exact same fields as ippc_channel_statistics
  char channel_id[CHANNEL_ID_MAX_LENGTH+1];

  int send_rate;
  int recv_rate;
  int allowed_send_rate;

  int loss_rate;
  int ave_rtt;
};

struct statistics_report { // same struct as ippcmsg_statistics
  int total_send_rate;
  int total_recv_rate;
  int total_allowed_send_rate;
  int ave_rtt; // smoothed RTT over all channels (weighted by allowed rate)
  int loss_rate; // loss rate over all channels.

  int nchannels;
  statistics_channel channel_stats[STAT_MAX_CHANNELS];
};


typedef void (*horde_callback_statistics_t)(struct statistics_report*);

void horde_statistics_set_callback(horde_callback_statistics_t cb);

#endif
