#include <signal.h>
#include <iostream>

#include "properties.h"

#include "daemon.h"
#include "net.h"

const char *DEFAULT_PROPERTIES_FILE = "horde.properties";
const char *DEFAULT_NAMESPACE_ROOT = "/var/horde";
const char* DEFAULT_INTERFACE_LIST = "eth0, eth1, ppp0, ppp1, ppp2, ppp3, tun0, tun1, tun2, tun3";
const char *DEFAULT_PRIMARY_PORT = "12345";
const char *DEFAULT_POOLMAN_SLEEP = "5000";

const char *DEFAULT_SESSION_DAEMON_COMMAND = "horded.session";
const char *DEFAULT_SESSION_DAEMON_WRAPPER = ""; // no wrapper
const char *DEFAULT_SESSION_DAEMON_OUT_REDIRECT = ""; // no redirect

io_event_loop io_loop;
string namespace_root;
int primary_port;

string session_daemon_command = DEFAULT_SESSION_DAEMON_COMMAND;
string session_daemon_wrapper;
string session_daemon_out_redirect;

string properties_filename = DEFAULT_PROPERTIES_FILE;

using namespace horde;

static void interrupt_signal(int s) {
  string sigstr;
  switch (s) {
  case SIGINT: sigstr = "SIGINT"; break;
  case SIGTERM: sigstr = "SIGTERM"; break;
  case SIGHUP: sigstr = "SIGHUP"; break;
  case SIGSEGV: sigstr = "SIGSEGV"; break;
  default: sigstr = "#" + s;
  }

  LOG_EVENT("signal") << "Received signal " << sigstr << ", exiting...";
  if (s == SIGSEGV) 
    horde::panic("segmentation fault.");
  else
    horde::exit();
}

// main function
main (int argc, char** argv) {
  // factor in current working directory into constants...
  char *cwd = get_current_dir_name();
  if (cwd) {
    session_daemon_command = cwd;
    session_daemon_command += '/';
    session_daemon_command += DEFAULT_SESSION_DAEMON_COMMAND;

    properties_filename = cwd;
    properties_filename += '/';
    properties_filename += DEFAULT_PROPERTIES_FILE;
    
    free(cwd);
  }

  //
  // handle command line arguments -- TODO: make this better?
  //  
  if (argc > 3) {
    cerr << "Correct usage:" << endl
	 << "\t" << argv[0] << "[properties_filename]";
  } else {
    if (argc > 1) properties_filename = argv[1];
  }

  //
  // load properties file
  //
  if (!load_properties_file(properties_filename)) {
    cerr << "Could not load properties file '" << properties_filename << "', using defaults..." << endl;
  }
  
  namespace_root = get_property("namespace_root", DEFAULT_NAMESPACE_ROOT);

  //
  // initialize logging
  //
  if (!horde::logging_init("MAIN", namespace_root.c_str())) {
    // critical error
    cerr << "CRITICAL ERROR: could not create log files. Namespace root: " << namespace_root << endl
	 << "Exiting." << endl;
    exit(-1);
  }

  //
  // start IPPC server and set command
  session_daemon_command = get_property("session_daemon_command", session_daemon_command);
  session_daemon_wrapper = get_property("session_daemon_wrapper", DEFAULT_SESSION_DAEMON_WRAPPER);
  session_daemon_out_redirect = get_property("session_daemon_out_redirect", DEFAULT_SESSION_DAEMON_OUT_REDIRECT);

  init_syscall(io_loop, namespace_root + "/sessions/ippc");

  //
  // start network I/O handlers
  //
  string iface_list = get_property("managed_interfaces", DEFAULT_INTERFACE_LIST);
  vector<string> all_ifaces = properties::decode_array_value(iface_list);
  set<string> unique_ifaces;
  for (vector<string>::iterator iter = all_ifaces.begin(); iter != all_ifaces.end(); iter++) {
    const char* iface = iter->c_str();
    if (iface[0] == '+') iface += 2;
    unique_ifaces.insert(string(iface));
  }

  vector<string> ifaces(unique_ifaces.begin(), unique_ifaces.end());

  primary_port = atoi(get_property("primary_port", DEFAULT_PRIMARY_PORT).c_str());

  int poolman_sleep_interval = atoi(get_property("poolman_sleep_interval", DEFAULT_POOLMAN_SLEEP).c_str());
  if (poolman_sleep_interval < 200) poolman_sleep_interval = 200;

  create_poolman(ifaces, primary_port, poolman_sleep_interval, io_loop);
  interface_pool_manager* poolman = get_poolman();

  if (poolman) {
    poolman->start();
  } else {
    shutdown_syscall();
    horde::panic("Could not create an interface pool manager.");
  }

  //
  // set up a periodic timer to cleanup zombie child processes
  //
  init_zombie_killer();

  // setup signal handlers
  signal(SIGINT, interrupt_signal);
  signal(SIGTERM, interrupt_signal);
  signal(SIGHUP, interrupt_signal);
  signal(SIGSEGV, interrupt_signal);
  
  siginterrupt(SIGINT, 1);
  siginterrupt(SIGTERM, 1);
  siginterrupt(SIGHUP, 1);
  siginterrupt(SIGSEGV, 1);

  //
  // startup message
  // 

  cout << "[horde:\033[1;31mdaemon\033[0m] started." << endl << flush; 

  //
  // enter I/O loop
  //
  io_loop.run();

  LOG_EVENT("i/o") << "unexpected exit from I/O loop.";
  shutdown_syscall();
  horde::exit();
}


