#include <sys/socket.h>
#include <net/if.h> 
#include <ifaddrs.h>
#include <string.h>

#include <nmstl/net>
#include <nmstl/ntime>

#include "daemon.h"
#include "net.h"
#include "packet_io.h"

//
// interface I/O manager
//
size_t RECV_BUFFER_SIZE = 1600;

void
horde::interface_iohandler::handle_packet_recv(horde_net_packet& packet, inet_address& addr)
{
  if ((packet.get_type() == PACKET_TYPE_SESSION_CONNECT_REQUEST) 
      || (packet.get_type() == PACKET_TYPE_SESSION_CONNECT_ACK)
      || (packet.get_type() == PACKET_TYPE_SESSION_CONNECT_NACK)) {

    string type="";
    
    // get session name from body of packet
    //size_t keysz = packet.get_data_size();
    char* key = packet.get_data();
    string session_name(key);

    // call the right handler function
    if (packet.get_type() == PACKET_TYPE_SESSION_CONNECT_REQUEST) {
      handle_recv_connect_request(addr, session_name);
    } else if (packet.get_type() == PACKET_TYPE_SESSION_CONNECT_ACK) {
      handle_recv_connect_ack(addr, session_name);
    } else {
      handle_recv_connect_nack(addr, session_name);
    }
    
    // log this recv
    LOG_NET("recv") << addr.as_string() << "\t" << packet
		    << "\tsession_name:\'" << session_name << "\'";
  }
  else { // unexpected packet type
    // log this error
    LOG_NET("error") << addr.as_string() << "\tbad packet: " << packet.to_debug_string();
  }
    
}

void horde::interface_iohandler::ravail() {
  // get the packet.
  horde_net_packet recv_buf(RECV_BUFFER_SIZE);
  inet_address recv_addr;
  
  if (recv_packet(recv_buf.get_buf(), recv_addr) < 0) {
    LOG_NET("error") << "recv_packet() failed: " << strerror(errno);
  } else {
    handle_packet_recv(recv_buf, recv_addr);
  }
}

void
horde::interface_iohandler::send_control_packet(inet_address peer_address, string session_name, horde_packet_type type) {
  size_t strlen = session_name.length();

  // construct the packet.
  horde_net_packet packet(horde_net_packet::header_sz + strlen, type);
  packet.set_length(horde_net_packet::header_sz + strlen);
  
  if (strlen > 0)
    memcpy(packet.get_data(), session_name.c_str(), strlen);
  
  // log packet for debugging purposes
  LOG_NET("send") << peer_address.as_string() 
		  << "\t" << packet << " \tsession_name:\'" << session_name << "\'";

  // send packet
  if (this->send_packet(packet.get_buf(), peer_address) < 0) {
    LOG_NET("error") << "send_packet() failed: " << strerror(errno);
  }
}

//
// pool manager
//
interface_pool_manager* the_poolman = NULL;

void 
horde::create_poolman(vector<string>& managed_ifaces, uint16_t port, 
		  uint32_t period, io_event_loop& loop) {
  if (the_poolman) {
    if (the_poolman->is_running()) {
      the_poolman->stop();
    }
    delete the_poolman;
  }

  the_poolman = new interface_pool_manager(managed_ifaces, port, period, loop);
}

interface_pool_manager* 
horde::get_poolman() {
  return the_poolman;
}

// --------- utility functions ----------

// bridge between nmstl and socket structs
// @param port: port to use, or if zero, port is read from addr
inet_address parse_sockaddr_in(sockaddr_in* addr, int port) { 
  char buf[INET_ADDRSTRLEN+1];  
  inet_ntop(AF_INET, &(addr->sin_addr), buf, INET_ADDRSTRLEN);

  if (port == 0) 
    port = ntohs(addr->sin_port);

  return inet_address(string(buf), port);
}

// looks up an inet4 interface.
// @return: true if an interface with the given name was found; false otherwise.
// @param addr: if the iface_name was found, copies contents of
//             address of first record that matched to this param.
// @param port: port to use, or if zero, port is read from addr
bool lookup_interface(string iface_name, inet_address& addr, int port = 0) {
  bool retval=false;

  const char *iface_id = iface_name.c_str();
  if (iface_id[0] == '+')  // simulated interface...
    iface_id = iface_id + 2;

  struct ifaddrs *iface, *ifaces;
  getifaddrs(&iface);
  ifaces = iface;

  for (; iface != NULL; iface = iface->ifa_next) {
    if ((iface->ifa_addr != NULL) && (iface->ifa_addr->sa_family == AF_INET)) {     

      if (strcmp(iface_id, iface->ifa_name) == 0) { // found it...
	// copy contents
	addr = parse_sockaddr_in((sockaddr_in*)iface->ifa_addr, port);
	retval=true; 
	break;
      }
    }
  }

  freeifaddrs(ifaces); // avoid mem leak.

  return retval;
}

// --------- start / stop --------------

void
horde::interface_pool_manager::start() {
  if (!running) {
    // debug
    string ifacestr("");
    for (vector<string>::iterator iter = ifaces.begin(); iter != ifaces.end(); iter++) {
      ifacestr += *iter;
      ifacestr += ' ';
    }

    LOG_EVENT("poolman") << "poolman started, monitoring: " << ifacestr;

    // start poolman
    running = true;
    wakeup();
  } // else: wakeup will be called soon, so do nothing...
}

void
horde::interface_pool_manager::stop() {

  // shut down every interface...
  if (running)
    for (int i = 0; i < ifaces.size(); i++)
      iface_down(i);

  LOG_EVENT("poolman") << "poolman stopped.";
  running = false;
}

// ---------- wakeup --------------------
void
horde::interface_pool_manager::wakeup() {
  if (running) {
    //LOG_NET("poolman") << "wakeup()";
    int idx = 0;
    for (vector<string>::const_iterator iter = ifaces.begin();
	 iter != ifaces.end(); iter++, idx++) {
      string iface = *iter;

      // find last round IP of this interface...
      STRING_HASHMAP(inet_address)::iterator if_addr = iface_addresses.find(iface);
      bool was_active = (if_addr != iface_addresses.end());
      
      // is interface up?
      inet_address address;
      if (lookup_interface(iface, address, port)) {
	// interface is currently up
	if (was_active) { // was this active in last round?
	  // we only care if the address has changed
	  if (if_addr->second != address) { 
	    LOG_EVENT("poolman") << "REBOOTING <" << iface << ">, address change.";
	    iface_down(idx);
	    iface_up(idx, address);
	  }
	} else {
	  // newly activated interface...
	  LOG_EVENT("poolman") << "UP <" << iface << ">, address:" << address << "port:" << port;
	  iface_up(idx, address);
	}
      } else {
	// interface is not currently up, but was active...
	if (was_active) {
	  LOG_EVENT("poolman") << "DOWN <" << iface << ">";
	  iface_down(idx); // active in last round, so clean up...
	}
      }     
    }

    // make sure logs on disk are up-to-date
    logging_flush();
    
    // reschedule wakeup...
    (new callback_timer
     (*ioloop, wrap(this, &horde::interface_pool_manager::wakeup)))->arm(ntime::now_plus_msecs(period_msecs));
  }
}

// ------------- iface_up / iface_down ---------------

// requires: # of iface_up calls - # of iface_down calls for this idx = 0
void
horde::interface_pool_manager::iface_up(int idx, inet_address address) {
  string iface = ifaces[idx];

  // add to set of active addresses
  iface_addresses[iface] = address;

  // create an iohandler for this interface
  ptr<interface_iohandler> ioh(new interface_iohandler(io_loop, address));
  
  iohandler_map[idx] = ioh;  
}

// requires: # of iface_up calls - # of iface_down calls for this idx = 1
void
horde::interface_pool_manager::iface_down(int idx) {
  string iface = ifaces[idx];

  // remove this interface from the set of iface_addresses set
  STRING_HASHMAP(inet_address)::iterator if_addr = iface_addresses.find(iface);
  if (if_addr != iface_addresses.end()) { // if found...
    iface_addresses.erase(if_addr); // remove from set of active addresses
  }

  // shut down the io handler for this interface
  hash_map<int, ptr<interface_iohandler> >::iterator iter = iohandler_map.find(idx);
  if (iter != iohandler_map.end()) {
    iter->second->shutdown(); // shut down io handler
    iohandler_map.erase(idx); // remove from table   
  }
}

// ---------------- packet I/O ------------------------

void 
horde::interface_pool_manager::send_control_packet(inet_address peer_address, 
						   string session_name,
						   horde_packet_type type) {
  // pick an interface
  for (int i = 0; i < ifaces.size(); i++) {
    hash_map<int, ptr<interface_iohandler> >::iterator iter = iohandler_map.find(i);
    if (iter != iohandler_map.end()) { // this interface is valid.

      ptr<interface_iohandler> ioh = iter->second;

      ioh->send_control_packet(peer_address, session_name, type); // send the request
      return; // sent a request, exit function.
    }
  }
}

void 
horde::interface_pool_manager::send_connect_request(inet_address peer_address, string session_name) {
  send_control_packet(peer_address, session_name, PACKET_TYPE_SESSION_CONNECT_REQUEST);
}

void 
horde::interface_pool_manager::send_connect_nack(inet_address peer_address, string session_name) {
  send_control_packet(peer_address, session_name, PACKET_TYPE_SESSION_CONNECT_NACK);
}
