#include "horde_log.h"
#include "session_daemon.h"

#include "event_handlers.h"
#include "session/session.h"
#include "core/control_netio.h"
#include "net/poolman.h"

#include <set>
#include <list>

void debug_touch_data(const char* buf, size_t length);

HORDE_NAMESPACE_BEGIN;

///@todo implement timeouts and resends...
namespace control_netio {
  string session_id;
  inet_address peer_address;

  set<inet_address> peer_address_set;

  //
  // session info
  //  
  void set_session_info(string req_session_id, inet_address req_peer_address) {
    session_id = req_session_id;
    peer_address = req_peer_address;
  }

  string get_session_id() {
    return session_id;
  }

  //
  // keep track of remote iface events
  //
  void handle_remote_iface_event(inet_address address, bool iface_up) {
    LOG_EVENT_DEBUG("net") << "remote iface event: up=" << iface_up << ", address=" << address;

    if (iface_up) {    // remote interface came up
      bool was_empty = peer_address_set.empty();
      peer_address_set.insert(address);
      if (was_empty) peer_address = address;      // this is the new peer address
    } else {          // remote interface went down                               
      peer_address_set.erase(address);
      if (!peer_address_set.empty()) {
	peer_address = *peer_address_set.begin(); // pick another peer address
      }
    }
  }

  //
  // composite control packets
  //     the idea here is to reduce the amount of overhead if, for example,
  //     16 streams connection requests are sent together.
  list<horde_net_packet> composite_sendq;
  size_t comp_sendq_size = 0;
  bool composite_resilient = false;
  callback_timer *composer = NULL;
  const int COMPOSITE_MAX_DELAY = 50; // milliseconds
  const int COMPOSITE_MAX_SIZE = 768; // the maximum composite packet size in bytes

  int get_max_send_delay() {
    return COMPOSITE_MAX_DELAY;
  }

  void send_composite_now() {
    size_t n = composite_sendq.size();
    if (n == 1) {        // only one packet, just send it...
      LOG_EVENT_DEBUG("send") << "composite control x" << n;
      
      get_poolman()->send_control_packet(peer_address, composite_sendq.front(), composite_resilient);
    } else if (n > 1) {  // multiple packets, pack them together
      LOG_EVENT_DEBUG("send") << "composite control x" << n;
      
      // invariant: comp_sendq_size < COMPOSITE_MAX_SIZE
      horde_net_packet packet(comp_sendq_size + horde_net_packet::header_sz, PACKET_TYPE_COMPOSITE_CONTROL);
      char *pdata = packet.get_data();      
      for (list<horde_net_packet>::iterator iter = composite_sendq.begin(); iter != composite_sendq.end(); iter++) {
	dynbuf buf = (*iter).get_buf();
	size_t len = buf.length();
	short *psize = (short*)pdata;
	*psize = len;
	pdata += sizeof(short);
	memcpy(pdata, buf.data(), len);
	pdata += len;
      }
      get_poolman()->send_control_packet(peer_address, packet, composite_resilient);
    }
    
    composite_resilient = false;
    composite_sendq.clear();
    comp_sendq_size = 0;
    composer = NULL;
  }

  void compose_and_send_control_packet(inet_address& peer_addr, horde_net_packet& packet, bool resilient) {
    composite_resilient |= resilient;
    composite_sendq.push_back(packet);
    comp_sendq_size += sizeof(short) + packet.get_length();    

    if (comp_sendq_size > COMPOSITE_MAX_SIZE) { // if enough data, send immediately
      LOG_EVENT_DEBUG("send.virt") << "too much composite control data to wait any longer...";
      send_composite_now();
    } else if (composer == NULL) { // otherwise create a timeout, if one doesn't already exist
      composer = (new callback_timer(io_loop, wrap(send_composite_now)));
      composer->arm(ntime::now_plus_msecs(COMPOSITE_MAX_DELAY));

      LOG_EVENT_DEBUG("send.virt") << "armed composite send timer " << COMPOSITE_MAX_DELAY;
    } else {
      //LOG_EVENT_DEBUG("send.virt") << "composite timer already armed: " << (void*) (composer);
    }
  }

  void recv_dispatcher(inet_address from, horde_packet_type type, const char* data, size_t len);

  void unpack_and_dispatch_composite(inet_address from, const char* data, size_t len) {
    const char* pdata = data;
    const char* end = data + len;
    short* psize;
    packet_header *header;
    horde_packet_type type;

    while (pdata < end) {
      psize = (short*)pdata;
      pdata += sizeof(short);
      header = (packet_header*)pdata;
      type = (horde_packet_type)(header->type & 0x0f);

      recv_dispatcher(from, type, pdata + horde_net_packet::header_sz, (size_t)*psize);

      pdata += (size_t)(*psize);
    }
  }

  //
  // control packet formats
  //
  #define SESSION_ID_MAX_LENGTH 63
  #define STREAM_ID_MAX_LENGTH 63
  #define IFACE_ADDRESS_MAX_LENGTH 16

  struct session_payload {
    char session_id[SESSION_ID_MAX_LENGTH+1];
  };

  struct stream_payload {
    char   stream_id[SESSION_ID_MAX_LENGTH+1];
    int8_t local_uid;
    int8_t remote_uid;
  };

  struct ifevent_payload {
    char address[IFACE_ADDRESS_MAX_LENGTH+1];
    short port;
    int8_t event_type;
  };

  void send_ifevent_control_packet(inet_address peer_addr, inet_address address, bool iface_up) {    
    // construct the packet.
    size_t n = horde_net_packet::header_sz + sizeof(struct ifevent_payload);
    horde_net_packet packet(n, PACKET_TYPE_INTERFACE_EVENT);
    packet.set_length(n);
   
    ifevent_payload* body = (ifevent_payload*)packet.get_data();
    bzero(body, sizeof(struct ifevent_payload));
    strncpy(body->address, address.get_host().c_str(), IFACE_ADDRESS_MAX_LENGTH);
    body->port = address.get_port();
    body->event_type = (iface_up) ? 1 : 0;
        
    // log packet for debugging purposes
    LOG_EVENT_DEBUG("send.virt") << peer_addr.as_string() 
				 << " <- PACKET_TYPE_INTERFACE_EVENT address: " << body->address
				 << ", port:" << body->port
				 << ", up: " << iface_up;
    
    // send packet
    compose_and_send_control_packet(peer_addr, packet, false);
  }

  void send_session_control_packet(inet_address peer_addr, string session_name, horde_packet_type type, bool resilient) {
    // construct the packet.
    size_t n = horde_net_packet::header_sz + sizeof(struct session_payload);
    horde_net_packet packet(n, type);
    packet.set_length(n);
   
    session_payload* body = (session_payload*)packet.get_data();
    bzero(body->session_id, SESSION_ID_MAX_LENGTH+1);
    strncpy(body->session_id, session_name.c_str(), SESSION_ID_MAX_LENGTH);
    
    // log packet for debugging purposes
    LOG_EVENT("send") << peer_addr.as_string() 
		      << " <- " << packet_type_to_string(type) << " name:\'" << session_name << "\'";
    
    // send packet
    get_poolman()->send_control_packet(peer_addr, packet, resilient, true);   
  }

  void send_stream_control_packet(inet_address peer_addr, string stream_id, int local_uid, int remote_uid, horde_packet_type type, bool resilient) {
    // construct the packet.
    size_t n = horde_net_packet::header_sz + sizeof(struct stream_payload);
    horde_net_packet packet(n, type);
    packet.set_length(n);
   
    stream_payload* body = (stream_payload*)packet.get_data();
    bzero(body->stream_id, STREAM_ID_MAX_LENGTH+1);
    strncpy(body->stream_id, stream_id.c_str(), STREAM_ID_MAX_LENGTH);
    body->local_uid = local_uid;
    body->remote_uid = remote_uid;
    
    // log packet for debugging purposes
    LOG_EVENT("send.virt") << peer_addr.as_string() 
			   << " <- " << packet_type_to_string(type) << " name:\'" << stream_id 
			   << "\', luid:" << local_uid << ", ruid:" << remote_uid;
    
    // send packet
    //get_poolman()->send_control_packet(peer_addr, packet, resilient);   
    compose_and_send_control_packet(peer_addr, packet, resilient);
  } 

  //
  // session control packets
  //
  void send_session_connect_request(string session_id, inet_address peer_address) {
    send_session_control_packet(peer_address, session_id, PACKET_TYPE_SESSION_CONNECT_REQUEST, false);
  }
  
  void send_session_connect_ack(string session_id, inet_address peer_address) {
    send_session_control_packet(peer_address, session_id, PACKET_TYPE_SESSION_CONNECT_ACK, true);
  }
  
  void send_session_close() {
    send_session_control_packet(peer_address, session_id, PACKET_TYPE_SESSION_CLOSE_REQUEST, true);
  }
  
  //
  // stream control packets
  //
  
  void send_stream_connect_request(string stream_id, int local_uid, int remote_uid) {
    send_stream_control_packet(peer_address, stream_id, local_uid, remote_uid, PACKET_TYPE_STREAM_CONNECT_REQUEST, false);
  }

  void send_stream_connect_ack(string stream_id, int local_uid, int remote_uid) {
    send_stream_control_packet(peer_address, stream_id, local_uid, remote_uid, PACKET_TYPE_STREAM_CONNECT_ACK, true);
  }

  void send_stream_close(string stream_id, int local_uid, int remote_uid) {
    send_stream_control_packet(peer_address, stream_id, local_uid, remote_uid, PACKET_TYPE_STREAM_CLOSE_REQUEST, true);
  }

  //
  // handle incoming packets
  //
  string extract_session_id(const char* data, size_t len) {
    return string(((session_payload*)data)->session_id);
  }

  // returns the stream_id and sets the local uid and remote uid
  void extract_stream_params(const char* data, size_t len, string& id, int& luid, int& ruid) {
    stream_payload* body = (stream_payload*)data;
    luid = body->local_uid;
    ruid = body->remote_uid;
    id = string(body->stream_id);
  }

  void extract_ifevent_params(const char* data, size_t len, inet_address& addr, bool& flag) {
    ifevent_payload* body = (ifevent_payload*)data;
    flag = (body->event_type == 0) ? false : true;
    addr = inet_address(body->address, body->port);
  }

  /// function that parses a packet and calls the right function to handle it (calls itself
  /// recursively in the case of COMPOSITE_CONTROL packets).
  void recv_dispatcher(inet_address from, horde_packet_type type, const char* data, size_t len) {
    int local_uid, remote_uid;
    string stream_id;

    LOG_EVENT_DEBUG("net") << "control packet fragment recv'd: " << packet_type_to_string(type) << ", len=" << len;

    switch (type) {
    case PACKET_TYPE_SESSION_CONNECT_REQUEST:      
      // we shouldn't actually get connect requests in the session daemon...
      // unless the ack was dropped...
      notify_session_connect_request(extract_session_id(data, len), from);
      break;

    case PACKET_TYPE_SESSION_CONNECT_ACK:
      notify_session_connect_ack(extract_session_id(data, len), from);
      break;

    case PACKET_TYPE_SESSION_CONNECT_NACK:
      // no NACK's in current implementation
      break;

    case PACKET_TYPE_SESSION_CLOSE_REQUEST:
      notify_session_close_request(extract_session_id(data, len));
      break;
      
    case PACKET_TYPE_STREAM_CONNECT_REQUEST:
      extract_stream_params(data, len, stream_id, local_uid, remote_uid);
      notify_stream_connect_request(stream_id, remote_uid, local_uid); // his remote uid is my local uid
      break;
    case PACKET_TYPE_STREAM_CONNECT_ACK:
      extract_stream_params(data, len, stream_id, local_uid, remote_uid);
      notify_stream_connect_ack(stream_id, remote_uid, local_uid); // his remote uid is my local uid
      break;

    case PACKET_TYPE_STREAM_CONNECT_NACK:
      // no NACK's in current implementation
      break;

    case PACKET_TYPE_STREAM_CLOSE_REQUEST:
      extract_stream_params(data, len, stream_id, local_uid, remote_uid);
      notify_stream_close_request(stream_id, remote_uid, local_uid); // his remote uid is my local uid
      break;

    case PACKET_TYPE_INTERFACE_EVENT:
      {
	inet_address addr;
	bool flag;
	extract_ifevent_params(data, len, addr, flag);
	handle_remote_iface_event(addr, flag);
	break;
      }
    case PACKET_TYPE_COMPOSITE_CONTROL:
      {
	unpack_and_dispatch_composite(from, data, len); // calls recv_dispatcher for each piece in this packet.
      }
    default:
      break;
    }
  }

  void handle_packet_recv(inet_address from, horde_net_packet packet) {
    recv_dispatcher(from, packet.get_type(), packet.get_data(), packet.get_data_size());
  }

};

inet_address get_peer_address() {
  return control_netio::peer_address;
}

void notify_core_iface_down(inet_address address) {
  if (session::is_connected()) {
    // notify peer
    control_netio::send_ifevent_control_packet(control_netio::peer_address, address, false);
  }
}

void notify_core_iface_up(inet_address address) {
  if (session::is_connected()) {
    // notify peer
    control_netio::send_ifevent_control_packet(control_netio::peer_address, address, true);
  }
}

void notify_core_session_connected() {
  get_poolman()->synchronize_core(); // force iface up events for every active interface
}


HORDE_NAMESPACE_END;
