/** @file control_netio.h 
 *
 * Functions to support network I/O for the control protocol.
 *
 */ 
#ifndef __HORDE_CONTROL_NETIO_H__
#define __HORDE_CONTROL_NETIO_H__

#include "horde_namespace.h"

#include <string>
using namespace std;

namespace nmstl {
  class inet_address;
};
using namespace nmstl;

HORDE_NAMESPACE_BEGIN;

/**
 * Gets a valid internet address for the network peer. The address returned by this function may
 * change during a session if the peer experiences an interface disconnection.
 */
inet_address get_peer_address();

void notify_core_iface_up(inet_address address); ///< called by poolman when iface comes up
void notify_core_iface_down(inet_address address); ///< called by poolman when iface goes down

namespace control_netio {

  //
  // session
  //

  /**
   * Sends a connection request to the specified peer.
   */
  void send_session_connect_request(string session_id, inet_address peer_address); 

  /**
   * Sends a connection request acknowledgement to the specified peer.
   */
  void send_session_connect_ack(string session_id, inet_address peer_address); 

  /**
   * Sets the info for the current session. 
   * 
   * The given \a session_id and \a peer_address will be used by all
   * network I/O functions to communicate.
   */
  void set_session_info(string session_id, inet_address peer_address);

  /**
   * Returns the \a session_id for the current session.
   */
  string get_session_id();   

  /**
   * Sends a 'close session' network control packet to the peer.
   *
   * @note This function blocks. The packet will be sent (or at least
   * in the OS kernel) by the time this call returns.
   */
  void send_session_close();

  //
  // streams
  //

  /**
   * Sends a stream connection request.
   */
  void send_stream_connect_request(string stream_id, int local_uid, int remote_uid); 

  /**
   * Sends a stream connection acknowledgement.
   */
  void send_stream_connect_ack(string stream_id, int local_uid, int remote_uid); 

  /**
   * Sends a 'close stream' network control packet.
   */
  void send_stream_close(string stream_id, int local_uid, int remote_uid);

  //
  // misc stuff
  //

  /** 
   * Gets the maximum number of milliseconds a send will be delayed
   * after a \a send_stream call. Sends may be delayed to pack
   * multiple send requests into a single packet.
   */
  int get_max_send_delay();
};

HORDE_NAMESPACE_END;

#endif
