// -*- mode: C++; -*-

/** @file scheduler.h
 *
 * Packet scheduler interface and declaration of the factory
 * function used to create scheduler objects.
 */
#ifndef __HORDE_SESSION_SCHEDULER_H__
#define __HORDE_SESSION_SCHEDULER_H__

#include <string>
#include <vector>
using namespace std;

#include <nmstl/ptr>
using namespace nmstl;

#include "horde_namespace.h"

HORDE_NAMESPACE_BEGIN;

class tx_slot;

/** 
 * Information on a fragment from an ADU.
 * 
 * This is a packed 80-bit data structure. It only supports a 24-bit
 * space of ADU identifiers, and ADU's smaller than 65536 bytes.
 */
class data_fragment_info {
 private:
  uint8_t  stream_id;    ///< stream identifier
  uint8_t  upper_adu_id; ///< top 8 bits of 24-bit adu identifier
  uint16_t lower_adu_id; ///< bottom 16 bits of adu identifier
  uint16_t frag_start;   ///< byte offset of this fragment in the given ADU
  uint16_t frag_len;     ///< length of this fragment, in bytes.
  uint16_t total_len;    ///< total length of the ADU, in bytes.

 public:
  data_fragment_info(uint8_t stream_id, int adu_id, uint16_t frag_start, uint16_t frag_len, uint16_t total_len)
    : stream_id(stream_id), frag_start(frag_start), frag_len(frag_len), total_len(total_len) {

    lower_adu_id = adu_id & 0x0000ffff;
    upper_adu_id = (adu_id & 0x00ff0000) >> 16;
  }

  /** Returns the 24-bit ADU identifier. */
  int get_adu_id() const {
    int id = lower_adu_id | (upper_adu_id << 16);
    return id;
  }

  int get_stream_id()  const { return stream_id; }
  int get_frag_start() const { return frag_start;}
  int get_frag_len()   const { return frag_len;  }
  int get_total_len()  const { return total_len; }

  int set_stream_id(uint8_t new_stream_id) { stream_id = new_stream_id; }
};

/**
 * A packet scheduler.
 *
 * The scheduler may pack one or more ADU fragments into a TX slot. Each
 * slot must consist of a sequence of one or more data fragments stored
 * as followed:
 *
 * \code
 * data_fragment_info fragment_header
 * char data[n]
 * \endcode
 *
 * The scheduler should generate unique \a packet_uid's for each
 * packet it produces, and it should keep track of which uid maps to
 * which set of fragments. It does not have to keep this mapping
 * around forever, but should cache it for long enough so that streams
 * can be appropriately notified of ACK and LOSS events.
 */
class scheduler {
 public:
  /**
   * Schedule a TX slot: pack the slot with data and consume it.
   *
   * This is called in response to a call to notify_core_new_tx_slot()
   * and should result in a call to consume_tx_slot() if some data is
   * ready to be sent, or this slot should be ignored. Queuing slots
   * for later use may result in unexpected behaviour.
   *
   */
  virtual void schedule_tx_slot(ptr<tx_slot> slot) = 0; 

  /**
   * Gets information on all fragments that were packed into a specified
   * packet, or returns an empty vector, if that information is either
   * no longer available, or the packet identifier is invalid.
   */
  virtual vector<data_fragment_info> get_packet_contents(int packet_uid) = 0;
};

/**
 * Creates a new scheduler object.
 *
 * See scheduler_impl.h for details on what kinds of scheduler types
 * are supported.
 *
 * @return NULL if unsupported \a type_id, otherwise a valid scheduler object.
 */
scheduler* create_scheduler(string type_id);

HORDE_NAMESPACE_END;

#endif
