#include <nmstl/net>

#include "horde_namespace.h"
#include "properties.h"
#include "horde_log.h"

#include "event_handlers.h"

#include "net/txslot.h"
#include "packet_io.h"

#include "core.h"
#include "scheduler.h"

using namespace horde;
using namespace nmstl;
using namespace std;

HORDE_NAMESPACE_BEGIN;

namespace control_netio { void handle_packet_recv(inet_address from, horde_net_packet packet); };
void unpack_and_deliver_data(horde_net_packet packet);

//
// variables used by core module
//
io_event_loop* core_ioloop = NULL;
scheduler*     active_scheduler = NULL;

int xmux_recv_count = 0; // shared variable

// 
// initialization
//

bool init_core(string scheduler_type, io_event_loop& ioloop) {
  core_ioloop = &ioloop;
  active_scheduler = create_scheduler(scheduler_type);

  LOG_EVENT_DEBUG("core") << "active_scheduler=" << (void*)active_scheduler;

  return (active_scheduler != NULL);
}

//
// callbacks invoked from network layer
//
void notify_core_new_tx_slot(ptr<tx_slot> slot) {
  if (active_scheduler) active_scheduler->schedule_tx_slot(slot);
}

void notify_core_ack_received(int packet_uid, int rtt) {
  if (active_scheduler) {
    LOG_EVENT_DEBUG("netack") << "ack for packet " << packet_uid;

    vector<data_fragment_info> frags = active_scheduler->get_packet_contents(packet_uid);
    for (vector<data_fragment_info>::const_iterator iter = frags.begin(); iter != frags.end(); iter++) {
      LOG_EVENT_DEBUG("netack") << "delivering ack stream_id=" << iter->get_stream_id() << ", adu=" << iter->get_adu_id() 
				<< ", offset=" << iter->get_frag_start() << ", len=" << iter->get_frag_len() << ", rtt=" << rtt;

      notify_stream_data_ack(iter->get_stream_id(), iter->get_adu_id(), iter->get_frag_start(), iter->get_frag_len(), rtt);
    }
  }
}

void notify_core_loss_detected(int packet_uid) {
  if (active_scheduler) {
    vector<data_fragment_info> frags = active_scheduler->get_packet_contents(packet_uid);
    for (vector<data_fragment_info>::const_iterator iter = frags.begin(); iter != frags.end(); iter++) {
      notify_stream_data_loss(iter->get_stream_id(), iter->get_adu_id(), iter->get_frag_start(), iter->get_frag_len());
    }
  }
}

string debugstr(char* data, int sz) {
  int len = sz;
  if (len > 24) len = 24;

  string str="'";
  char buf[8];

  unsigned char *pch = (unsigned char *)data;
  unsigned char *end = pch + len;

  for (; pch < end; pch++) {
    if ((*pch > 19) && (*pch < 128)) {
      str += *pch;
    } else {
      str += "?";
    }
  }

//   bool lastch = true;
//   for (; pch < end; pch++) {
//     if ((*pch > 79) || (*pch < 20) || (*pch == '|')) {
//       snprintf(buf,8,"%02x", (int)*pch);
//       if (lastch) str += "|";
//       str += buf;
//       lastch = false;
//     } else {
//       if (!lastch) str += "|";
//       str += *pch;      
//       lastch = true;
//     }
//   }
//   if (!lastch) str += "|";
//   if (len < sz) str += "...";
   str += "'";
  return str;
}

void notify_core_packet_received(horde_net_packet packet, inet_address from_address, int channel_id) {
  LOG_EVENT_DEBUG("netrecv") << "packet received, type=" << packet_type_to_string(packet.get_type())
			     << ", channel_seq=" << packet.get_header()->get_sequence();

  xmux_recv_count++;

  switch (packet.get_type()) {
  case PACKET_TYPE_DATA:
    // data packet (unpack the data packet and hand each fragment to the owner stream.)
    LOG_EVENT_DEBUG("netrecv") << "data: " << debugstr(packet.get_data(), packet.get_data_size());
    unpack_and_deliver_data(packet);
    break;
  case PACKET_TYPE_SESSION_CONNECT_REQUEST:
  case PACKET_TYPE_SESSION_CONNECT_ACK:
  case PACKET_TYPE_SESSION_CONNECT_ACK_ACK:
  case PACKET_TYPE_SESSION_CONNECT_NACK:
  case PACKET_TYPE_SESSION_CLOSE_REQUEST:
  case PACKET_TYPE_STREAM_CONNECT_REQUEST:
  case PACKET_TYPE_STREAM_CONNECT_ACK:
  case PACKET_TYPE_STREAM_CONNECT_NACK:
  case PACKET_TYPE_STREAM_CLOSE_REQUEST:
  case PACKET_TYPE_INTERFACE_EVENT:
  case PACKET_TYPE_COMPOSITE_CONTROL:
    // control packet (hand it to the control_netio module)
    control_netio::handle_packet_recv(from_address, packet);
    break;
  default:
    LOG_EVENT("net") << "unexpected packet type received: " << (int)packet.get_type();
  }
}

//
// helper functions
//

void unpack_and_deliver_data(horde_net_packet packet) {
  char *pdata = packet.get_data();
  char *end = pdata + packet.get_data_size();
  size_t offs = 0;
  
  while (pdata < end) {
    data_fragment_info* header = (data_fragment_info*)pdata;
    pdata += sizeof (struct data_fragment_info);
    offs += sizeof (struct data_fragment_info);

    LOG_EVENT_DEBUG("netrecv") << "delivering fragment: stream=" << header->get_stream_id() 
			       << ", adu=" << header->get_adu_id() << ", offset=" << header->get_frag_start()
			       << ", len=" << header->get_frag_len() << ", totlen=" << header->get_total_len()
			       << ", packet-offset=" << offs;

    notify_stream_data_recv(header->get_stream_id(), header->get_adu_id(),
			    header->get_frag_start(), header->get_frag_len(), header->get_total_len(),
			    pdata);
    pdata += header->get_frag_len();
    offs += header->get_frag_len();
  }
}

HORDE_NAMESPACE_END;
