#include <math.h>
#include "channel_sim.h"

HORDE_NAMESPACE_BEGIN;

/// Return a random number between 0.0-1.0
double ranf() {
  return ((double)rand())/RAND_MAX;
}

/**
 * Normal random variate generator. 
 *
 * from: ftp://ftp.taygeta.com/pub/c/boxmuller.c
 *
 * @param m mean
 * @param s standard deviation.
 */ 
double box_muller(double m, double s)
{
	double x1, x2, w, y1;
	static float y2;
	static int use_last = 0;

	if (use_last)		        /* use value from previous call */
	{
		y1 = y2;
		use_last = 0;
	}
	else
	{
		do {
			x1 = 2.0 * ranf() - 1.0;
			x2 = 2.0 * ranf() - 1.0;
			w = x1 * x1 + x2 * x2;
		} while ( w >= 1.0 );

		w = sqrt( (-2.0 * log( w ) ) / w );
		y1 = x1 * w;
		y2 = x2 * w;
		use_last = 1;
	}

	return( m + y1 * s );
}

/// Return a random value, using a Gaussian random variable with mean \a m and standard deviation \a sdev.
int rand_gaussian(int m, int sdev) {
  int r=(int)box_muller((double)m, (double)sdev);
  return r;
}

/// Return an approximately exponentially distributed number between 0.0 and 1.0
double rand_exponential(double lambda) {
  if (lambda > 0) {
    double result = (-1 * log(1.0 - ranf())) / lambda;
    return result;
  } else {
    return 0.0;
  }
}

simulated_channel_iohandler::simulated_channel_iohandler
                                  (properties& params,
				   io_event_loop& ioloop, string channel_name, inet_address bind_addr,
				   callback<void, horde_net_packet, inet_address> recv_handler,
				   callback<void> stat_handler, size_t recv_buffer_sz) 
				    : last_send(ntime::now()), 
				      channel_iohandler(ioloop, channel_name, bind_addr, recv_handler, stat_handler, recv_buffer_sz)
{
  // use params to init simulation model...
  // params:
  //
  //    sim.rtt.mean   int   : milliseconds
  //    sim.rtt.sdev   int
  //    sim.loss       float : random loss rate (0.0-1.0)
  //
  rtt_mean = atoi(params.get("sim.rtt.mean", "100").c_str());    // default: 100 ms
  if (rtt_mean < 0) rtt_mean = 0;

  rtt_sdev = atoi(params.get("sim.rtt.sdev", "0").c_str());      // default: no var
  if (rtt_sdev < 0) rtt_sdev = 0;
  if (rtt_sdev > rtt_mean*2) rtt_sdev = rtt_mean*2;

  loss = atof(params.get("sim.loss", "0.0").c_str());            // default: no loss
  if (loss > 0.95) loss = 0.95;
  if (loss < 0.00) loss = 0.00;

  LOG_EVENT("simulate") << "created simulated_channel_iohandler on " << bind_addr  << " [ rtt_mean=" 
			<< rtt_mean << ", sdev=" << rtt_sdev << ", loss=" << loss << ',' << ']';

  // init other vars 
  last_send = ntime::now();    
}

void 
simulated_channel_iohandler::simulated_send(horde_net_packet buffer, inet_address address) {
  ntime now = ntime::now();

  double random = ranf();
  if (random < loss) { // simulate random loss...
    // drop packet, but still count the bytes
    count_sent_bytes += buffer.get_length() - horde_net_packet::header_sz; 

    last_send = now;

    LOG_NET("sim.loss") << float_to_str(random, "%0.3lf") << "\t" 
			<< float_to_str(loss, "%0.3lf") << "\t" << buffer;
  } else {
    int rtt = rand_gaussian(rtt_mean, rtt_sdev);   // pick a random latency for this packet
    
    ntime next_send = now + ntime::msecs(rtt);
    if (next_send <= last_send) next_send = last_send + ntime::msecs(1); // don't reorder    

    rtt = (next_send - now).to_msecs();

    if (rtt > 0) {
      LOG_NET("sim.net") << "simulated rtt=" << rtt;

      // set up a delayed send, using a timer
      callback_timer* timer = (new callback_timer
			       (*(get_loop()), wrap(this, &horde::simulated_channel_iohandler::send_now, buffer, address)));
      timer->arm(next_send); // simulate latency
    } else {
      send_now(buffer, address);
    }
    
    last_send = next_send;
  }
}

HORDE_NAMESPACE_END;
