// -*- mode: C++; -*-

#ifndef __HORDE_DAEMON_NET_POOLMAN_H__
#define __HORDE_DAEMON_NET_POOLMAN_H__

#include <nmstl/netioevent>

#include "horde_namespace.h"
#include "string_hashmap.h"
#include "packet_io.h" // argh, better way?
#include "manager.h"

HORDE_NAMESPACE_BEGIN;

/**
 * The network channel pool manager. Periodically monitors the state of network interfaces,
 * and creates or destroys channel managers when appropriate.
 */
class network_pool_manager {
private:
  /// the scheduling period
  int period_msecs;
    
  /// is this poolman running?
  bool running;
    
  /// list of managed interfaces
  vector<string> ifaces;

  /// manager types for each managed interface
  STRING_HASHMAP(string) iface_manager_types;   
    
  /// interface information snapshot
  STRING_HASHMAP(inet_address) iface_addresses;

  /// the I/O loop to use
  io_event_loop* ioloop;

  /// interace index -> channel I/O handler map
  hash_map<int, ptr<network_channel_manager> > iohandler_map;
  
  //
  // internal functions to handle pool management...
  //
    
  /// wakeup, examine state of interfaces, determine what needs to be done and do it...
  void wakeup();
    
  /// note that an interface came online, init the io handlers
  void iface_up (int idx, inet_address addr);
    
  /// note that an interface went off, destroy the io handlers
  void iface_down(int idx);
    
  /**
   * Creates, but does not start, a network pool manager.
   *
   * @param period Number of milliseconds between wakeup()'s.
   * @param managed_ifaces A list of interface names (pair::first) and manager types (pair::second).  
   */
  network_pool_manager(vector< pair<string, string> >& managed_ifaces, int period, io_event_loop& loop);

  /// Returns NULL or a channel manager for the \a channel_id
  ptr<network_channel_manager> get_manager_from_id(int channel_id);

public:
  bool is_running() { return running; }
      
  /// Starts pool manager, unless is_running().
  void start();  
      
  /// Stops pool mnager, and deactivates all channel managers and io handlers.
  void stop();

  /**
   * Picks a channel and sends a control packet over that channel. If
   * resilient is true, this sends duplicates of the control packet on
   * \b all active channels.
   */
  void send_control_packet(inet_address peer_address, horde_net_packet packet, 
			   bool resilient=false, bool no_delay=false);

  /**
   * Get a list of objects that allow us to observe the states of
   * all the active network channels.
   */
  vector< ptr<channel_statistics> > get_statistics();

  /**
   * Calls notify_core_iface_up once for every active interface.
   */
  void synchronize_core();

  /**
   * Creates the pool manager.
   *
   * @param period Number of milliseconds between wakeup()'s.
   * @param managed_ifaces A list of interface names (pair::first) and manager types (pair::second).  
   */
  static void create_poolman(vector< pair<string, string> >& managed_ifaces, int period, io_event_loop& loop);

  /**
   * Returns the pool manager if it was successfully created, or NULL
   * if there was an error.
   */
  static network_pool_manager* get_poolman();
};

//
// exported functions
//
inline network_pool_manager* get_poolman() {
  return network_pool_manager::get_poolman();
}

inline void create_poolman(vector< pair<string, string> >& managed_ifaces, int period, io_event_loop& loop) {
  network_pool_manager::create_poolman(managed_ifaces, period, loop);
}

HORDE_NAMESPACE_END;

#endif
