// -*- mode: C++; -*-

/** @file session.h
 *
 * Session abstraction.
 *
 * @todo maybe closing a session should send a 'session closed' event to
 *       every active ippc_handler instead of just notifying the 
 *       session owner.
 */

#ifndef __HORDE_SESSION_H__
#define __HORDE_SESSION_H__

#include "horde_namespace.h"
#include "syscall/syscall.h" // for io_principal only

#include <string>
using namespace std;

HORDE_NAMESPACE_BEGIN;

/**
 * Initialize the session module. Register event handler hooks etc.
 */
void init_session();

namespace session {
  /// returns true if the session is in the connected state.
  bool is_connected();
  
  /**
   * Initializes the session.
   *
   * If \a acceptor is \a true, the session initialization involves
   * replying to the connection request packet that was received by
   * the identified peer. Otherwise, initialization involves sending
   * out a connection request and waiting for an acknowledgement.
   *
   * Once the session has been initialized, the \a owner is sent
   * either a 'session connected' event, or an exception.
   *
   * @param peer_address The internet address of the network peer (IP and port required).
   * @param owner The owner of this session. The principal where all session events will be sent.
   *              Only the owner is allowed to call \a close().
   */
  void initialize(string session_id, string peer_address, bool acceptor, io_principal owner);

  /**
   * Returns the session owner, if the session has been established, or NULL.
   */
  io_principal get_owner();

  /**
   * Returns the session identifier, if the session has been established, or "".
   */
  string get_id();

  /**
   * Allows an I/O principal to join a session. 
   * 
   * The principal will be sent a 'session connected' event if the
   * join is successful, or an exception. The principal will never be
   * sent another session related event.
   *
   * A join will only succeed, if the session is already 'connected'.
   */
  void join(string session_id, io_principal principal);

  /**
   * Closes a session. 
   *
   * If this succeeds, a 'session closed' event is delivered to the
   * caller; on failure an exception is delivered. Only succeeds if
   * the \a caller is the session owner.
   */
  void close(string session_id, io_principal caller);
};

HORDE_NAMESPACE_END;

#endif
