// -*- mode: C++; -*-

/** @file stream.h
 *
 * Stream abstraction.
 *
 */

#ifndef __HORDE_STREAM_H__
#define __HORDE_STREAM_H__

#include "horde_namespace.h"
#include "syscall/syscall.h"
#include "shared/objectives.h"

#include <list>
#include <string>
using namespace std;

#include <nmstl/io>
using namespace nmstl;

HORDE_NAMESPACE_BEGIN;

class stream {
protected:
  // --- send queue
  struct sendq_elem_t { 
    int adu_id;
    dynbuf data;
    int current_offset;

    bool is_fragment;       /// is this only part of an ADU?
    size_t adu_len;
    int adu_offset;

    horde_objective_tags tags;

    sendq_elem_t(int adu_id, char* buf, size_t buflen, 
		 bool is_frag, int offset, size_t total_len,
		 horde_objective_tags tags) 
      : adu_id(adu_id), data(buflen), current_offset(0),
	is_fragment(is_frag), adu_len(total_len), adu_offset(offset), tags(tags) 
    {
      data.set_length(buflen);
      memcpy(data.data(), buf, buflen);
    }
  };

  list<sendq_elem_t> sendq;
  size_t sendq_size;
  bool source_choked;

  static const int SENDQ_CHOKE_THRESHOLD = 8192;
  static const int SENDQ_MAX_SIZE = 307200; //= 16384;
  
  void sendq_push(sendq_elem_t& e);
  void sendq_pop();

  // ---

  enum stream_state_t {
    CLOSED,
    READY_TO_ACCEPT,
    TRYING_TO_CONNECT,
    CONNECTED
  } stream_state;         ///< The connection state of the stream.

  string stream_id;       ///< String ID (same on both peers)
  int stream_uid;         ///< Local UID.
  int remote_stream_uid;  ///< UID on remote peer.
  io_principal owner;     ///< owner of this stream.

  static const int EVENT_MASK_RECV = 0x0001;
  static const int EVENT_MASK_ACK = 0x0002;
  static const int EVENT_MASK_LOSS = 0x0004;
  static const int EVENT_MASK_FLUSHED = 0x0008;
  static const int EVENT_MASK_CHOKE = 0x00010;
  static const int EVENT_MASK_ALL = 0xffff;
  static const int EVENT_MASK_DEFAULT = EVENT_MASK_RECV | EVENT_MASK_ACK | EVENT_MASK_LOSS | EVENT_MASK_CHOKE;
    //EVENT_MASK_ALL ^ EVENT_MASK_FLUSHED;

  int event_mask;         ///< info on which events should be ignored

  /// construct a stream object.
  stream(stream_state_t state, string id, int uid, int remote_uid, io_principal owner) 
    : stream_state(state), stream_id(id), stream_uid(uid), remote_stream_uid(remote_uid),
      event_mask(EVENT_MASK_DEFAULT), owner(owner), sendq(), sendq_size(0), source_choked(false)
  { }

  // uid stuff

  /** 
   * Allocates a stream UID that is locally unique among all active streams.
   *
   * @return a number between 1-255 or 0 if no uid's are free 
   */
  static int uid_alloc();

  /// free's the given uid.
  static void uid_free(int uid);

  static void repeated_send_connect_request(stream* pstream);

  /// called when stream wants to enter the CONNECTED state.
  void connected(int remote_uid);

  // callbacks
  friend void notify_stream_connect_request(string session_id, int local_uid, int remote_uid);
  friend void notify_stream_connect_ack(string stream_id, int local_uid, int remote_uid);
  friend void notify_stream_close_request(string stream_id, int local_uid, int remote_uid);
  friend void notify_stream_data_ack(int stream_id, int adu_id, int frag_start, int frag_len, int rtt);
  friend void notify_stream_data_loss(int stream_id, int adu_id, int frag_start, int frag_len);
  friend void notify_stream_data_recv(int stream_id, int adu_id, int frag_start, int frag_len, int tot_len, const char* data);
public:
  // getters
  string get_id()  { 
    return stream_id;  
  }

  int get_uid() { 
    return stream_uid; 
  }

  int get_remote_uid() {
    return remote_stream_uid;
  }

  bool is_open() {
    return (stream_state != CLOSED);
  }

  bool is_connected() {
    return (stream_state == CONNECTED);
  }
    
  string to_debug_string();   

  //
  // set stream options
  //

  /**
   * Sets which events will be sent to the stream owner.
   *
   * May result in an exception (e.g. when the \a caller is not the owner).
   */
  void set_event_mask(bool event_recv, bool event_ack, 
		      bool event_loss, bool event_choke, bool event_flushed,
		      io_principal caller);

  //
  // Output
  //

  ///@todo a better signature for send()?
  void send_adu(int adu_id, char* data, size_t data_length, 
		horde_objective_tags tags, io_principal caller);

  void send_fragment(int adu_id, int adu_offset, size_t adu_length,
		     char* fragment, size_t fragment_length, 
		     horde_objective_tags tags, io_principal caller);

  void sendq_clear(io_principal caller);

  //
  // Functions used by scheduler.
  //

  /// Tests if there is any data ready to be sent from this stream.
  bool sendq_empty() { return sendq.empty(); }

  /**
   * Reads the meta-data for the ADU at the head of the send queue.
   *
   * Any arguments that are NULL are ignored. All others are filled in.
   *
   * @pre sendq_empty() == false
   * @todo should this return a structure instead of having this unweildly 
   *       signature.
   */
  void sendq_read_metadata(int *adu_id, int *current_offset, int *length,
			   bool *is_fragment, int *adu_offset, int *adu_length,
			   horde_objective_tags *tags);

  /**
   * Reads some data from the ADU at the head of the send queue, 
   * and copies to the given data buffer.
   *
   * If data is copied, but some bytes from the current ADU remain,
   * the \a current_offset of the sendq head ADU is incremented. If
   * there are no more bytes in that ADU, it is removed from the sendq.
   *
   * @param max_read The maximum number of bytes to read from the ADU.
   * 
   * @return The number of bytes copied into \a dest_buf <= max_read. 
   *         Returns -1 on error (e.g., if there is nothing in the sendq). 
   *         Otherwise, if this returns < max_read, the ADU at the head 
   *         of the sendq has been removed. May return 0, for 0-length
   *         ADU's.
   *
   * @pre sendq_empty() == false
   */
  int sendq_read_data(char* dest_buf, size_t max_read);

  /**
   * Returns the ADU at the front of the send queue, or an empty buffer if
   * the queue is empty.
   */
  dynbuf sendq_front(int *adu_id);

  //
  // static methods
  //
    
  /** 
   * Returns a list of all the active streams. 
   *
   * @warning do not mutate the returned vector.
   */
  static vector<stream*> get_all_streams();

  /** 
   * Returns a list of all the streams in the 'connected' state.
   *
   * @warning do not mutate the returned vector.
   */
  static vector<stream*> get_connected_streams();

  /** Returns \a NULL, or an \a x such that <tt>(x->get_id() == id)</tt>. */
  static stream* lookup_stream(string id);

  /** Returns \a NULL, or an \a x such that <tt>(x->get_uid() == uid)</tt>. */
  static stream* lookup_stream(int uid);

  /**
   * Closes the specified stream, if it exists and is connected, and
   * deallocates any local data structures tied to this stream.
   *
   * @param notify_peer Whether or not a 'stream closed' notification
   *                    should be sent to the network peer.
   */
  static void close(string stream_id, io_principal caller, bool notify_peer=true);

  /**
   * Attempts to accept a stream connection.
   *
   * Waits for a remote stream connection request with the given \a
   * stream_id. If such a request arrives, the \a caller is notified
   * with a 'stream connected' event. The \a caller becomes the owner
   * of the newly created stream and all stream events are redirected
   * to this \a caller. If no connection request arrives, an event may
   * never be generated.
   * 
   * On an error, an exception is delivered to the caller.
   */
  static void accept(string stream_id, io_principal caller);
    
  /**
   * Attempts to open a stream connection.
   * 
   * Sends a connection request to the remote session peer, with the
   * given \a stream_id. If the request is acknowledged, the stream is
   * considered to be open and a 'stream connected' event is delivered
   * to the \a caller. This \a caller becomes the owner of the stream,
   * and all subsequent stream events are sent to it. If no
   * acknowledgement arrives, no event may ever be generated (there
   * are no internal timeouts that cancel a connect; horde may try 
   * to keep resending the connection request forever).
   *
   * On an error, an exception is delivered to the caller.
   */
  static void connect(string stream_id, io_principal caller);   

  /**
   * Closes any stream owned by \a caller.
   */
  static void close_all_streams_owned_by(io_principal caller, bool notify_peer=true);
};

HORDE_NAMESPACE_END;

#endif
