/// @todo verify ippc_handler pointers before calling their methods...

#include <nmstl/ioevent>
#include <unistd.h>

#include "ippc/ippc_server.h"
#include "ippc/syscall.h"
#include "syscall/syscall.h"
#include "horde_log.h"
#include "event_handlers.h"

#include "session/session.h"
#include "session/stream.h"

#include "core/stats.h"

using namespace horde;
using namespace nmstl;

HORDE_NAMESPACE_BEGIN;

ippc_server*  server = NULL;

void event_ippc_recv(ippc_handler* ioh, int type,char* msg, size_t len);
void event_ippc_close(ippc_handler* ioh);
void event_ippc_owner_flush(ippc_handler* ioh);

void send_session_closed_ippc(bool immediate);

ippc_handler*
init_syscall(io_event_loop& ioloop, string path, int owner_fd) {
  server = new ippc_server(ioloop, path, event_ippc_recv, event_ippc_close);
  ippc_handler* owner = server->create_ippc_handler(owner_fd, event_ippc_recv, event_ippc_close, event_ippc_owner_flush);

  return owner;
}

void send_exception(ippc_handler* ioh, ippc_exception_code excode, const char* msg) {
  ippcmsg_exception ex;
  bzero(&ex, sizeof ex);
  
  ex.code = excode;
  strncpy(ex.msg, msg, EXCEPT_MSG_MAX_LENGTH);

  ioh->send_message(IPPC_EXCEPTION, &ex, sizeof ex);
}

//
// syscall handlers
//

// IPPC_SESSION_COMMAND: joins, closes and statistics commands
void syscall_session_command(ippc_handler* ioh, char* msg, size_t len) {
  ippcmsg_session_command* com = (ippcmsg_session_command*)msg;

  /// @todo verify session_id (it's a string; and test that it is a valid name)
  
  // try to execute the command
  if (com->command == ippcmsg_session_command::JOIN) {
    // todo... session must be connected...
    // send a session_connected event to ioh...
      // execute command
    session::join(string(com->session_id), ioh);
  } else if (com->command == ippcmsg_session_command::CLOSE) {
    // log
    LOG_EVENT("syscall") << "IPPC_SESSION_COMMAND: command=CLOSE, session_id='"  << com->session_id << "'";
      
    // execute command
    session::close(string(com->session_id), ioh);
  } else if (com->command == ippcmsg_session_command::ENABLE_STATISTICS) {
    LOG_EVENT("syscall") << "IPPC_SESSION_COMMAND: command=ENABLE_STATISTICS";
    add_statistics_listener(ioh);
  } else if (com->command == ippcmsg_session_command::DISABLE_STATISTICS) {
    LOG_EVENT("syscall") << "IPPC_SESSION_COMMAND: command=DISABLE_STATISTICS";
    remove_statistics_listener(ioh);
  } else {
    // log the bad ippc code
    LOG_EVENT("syscall") << "Bad IPPC_SESSION_COMMAND code: " << (int)com->command;
    
    // reply with an exception    
    send_exception(ioh, IPPC_EX_BAD_IPPC, "Unsupported session command (bad command code in IPPC_SESSION_COMMAND).");
  }  
}

// IPPC_STREAM_COMMAND: connect/accept/close
void syscall_stream_command(ippc_handler* ioh, char* msg, size_t len) {  
  ippcmsg_stream_command* com = (ippcmsg_stream_command*)msg;

  /// @todo verify stream_id (it's a string; and test that it is a valid name)
  
  // try to execute the command
  if (com->command == ippcmsg_stream_command::CONNECT) {
    // log
    LOG_EVENT("syscall") << "IPPC_STREAM_COMMAND: command=CONNECT, stream_id='" << com->stream_id << "'";
    
    // execute command
    stream::connect(string(com->stream_id), ioh);
    
  } else if (com->command == ippcmsg_stream_command::ACCEPT) {
    // log
    LOG_EVENT("syscall") << "IPPC_STREAM_COMMAND: command=ACCEPT, stream_id='" << com->stream_id << "'";
    
    // execute command
    stream::accept(string(com->stream_id), ioh);
    
  } else if (com->command == ippcmsg_stream_command::CLOSE) {   
    // log
    LOG_EVENT("syscall") << "IPPC_STREAM_COMMAND: command=CLOSE, stream_id='" << com->stream_id << "'";
    
    // execute
    stream::close(string(com->stream_id), ioh);

  } else if (com->command == ippcmsg_stream_command::SET_EVENT_MASK) {   
    // log
    LOG_EVENT("syscall") << "IPPC_STREAM_COMMAND: command=SET_EVENT_MASK, stream_uid='" 
			 << com->stream_uid << "', ack=" << com->event_mask_ack
			 << ", loss=" << com->event_mask_loss << ", recv=" << com->event_mask_recv
			 << ", choke=" << com->event_mask_choke << ", flushed=" << com->event_mask_flushed;
    // execute
    stream* pstream = stream::lookup_stream(com->stream_uid);
    if (pstream) {
      pstream->set_event_mask(bool(com->event_mask_recv), bool(com->event_mask_ack),  bool(com->event_mask_loss), 
			      bool(com->event_mask_choke), bool(com->event_mask_flushed), ioh);
    } else {
      send_exception(ioh, IPPC_EX_BAD_STREAM_ID, "Invalid stream handle.");
    }
  } else if (com->command == ippcmsg_stream_command::CLEAR) {   
    // log
    LOG_EVENT("syscall") << "IPPC_STREAM_COMMAND: command=CLEAR, stream_uid='" << com->stream_uid << "'";

    // execute
    stream* pstream = stream::lookup_stream(com->stream_uid);
    if (pstream) {
      pstream->sendq_clear(ioh);
    } else {
      send_exception(ioh, IPPC_EX_BAD_STREAM_ID, "Invalid stream handle.");
    }
  } else {
    // log the bad ippc code
    LOG_EVENT("syscall") << "Bad IPPC_STREAM_COMMAND code: " << (int)com->command;
    
    // reply with an exception
    send_exception(ioh, IPPC_EX_BAD_IPPC, "Unsupported stream command (bad command code in IPPC_STREAM_COMMAND).");
  } 
}

string debugstr(char* data, int sz);

void syscall_data_send(ippc_handler* ioh, char* msg, size_t len) {
  ippcmsg_data_send* send = (ippcmsg_data_send*)msg;

  stream* stream = stream::lookup_stream(send->hstream);

  if (stream) {
    // verify objectives
    if (send->tags.loss_threshold < 0) send->tags.loss_threshold = 0;
    else if (send->tags.loss_threshold > 100) send->tags.loss_threshold =  100;

    if (send->tags.rtt_threshold < 0) send->tags.rtt_threshold = 0;

    // log
    LOG_EVENT("syscall") << "IPPC_DATA_SEND: stream=" << send->hstream 
			 << ", adu=" << send->adu_id 
			 << ", len=" << send->data_length
			 << ", fragment:" << send->is_fragment 
			 << ", offs=" << send->adu_offset
			 << ", totlen=" << send->adu_length;

    //LOG_EVENT_DEBUG("debug") << "data: " << debugstr(send->data, send->data_length);

    if (send->is_fragment)
      stream->send_fragment(send->adu_id, send->adu_offset, send->adu_length,
			    send->data, send->data_length, send->tags, ioh);
    else 
      stream->send_adu(send->adu_id, send->data, send->data_length, send->tags, ioh);

  } else {
    LOG_EVENT("syscall") << "IPPC_DATA_SEND: unresolvable stream handle=" << send->hstream;
    LOG_EVENT_DEBUG("syscall") << len << "-byte message: " << to_hexstr(msg, len);
    send_exception(ioh, IPPC_EX_BAD_STREAM_ID, "Invalid stream handle.");
  }
}

// NOP, for debugging
void syscall_nop(ippc_handler* ioh, char* msg, size_t len) {
  // do nothing
  LOG_EVENT_DEBUG("syscall") << "NOP";
}

// PING, for debugging, and can be used to test is if daemon is working...
void syscall_ping(ippc_handler* ioh, char* msg, size_t len) {
  ippcmsg_pingpong* ping = (ippcmsg_pingpong*)msg;

  LOG_EVENT_DEBUG("syscall") << "Ping " << ping->seq; // log

  ioh->send_message(IPPC_PONG, ping, sizeof(struct ippcmsg_pingpong)); // echo
}

//
// ippc_handler event handlers
//

// handle an IPPC recv (dispatch to appropriate syscall_* function)
void event_ippc_recv(ippc_handler* ioh, int type, char* msg, size_t len) {
  switch (type) {
  case IPPC_SESSION_COMMAND:
    syscall_session_command(ioh, msg, len);
    break;
  case IPPC_STREAM_COMMAND:
    syscall_stream_command(ioh, msg, len);
    break;
  case IPPC_DATA_SEND:
    syscall_data_send(ioh, msg, len);
    break;
  case IPPC_PING:
    syscall_ping(ioh, msg, len);
    break;
  case IPPC_NOP:
    syscall_nop(ioh, msg, len);
    break;
  default:
    // log the bad ippc codea
    LOG_EVENT("syscall") << "Bad IPPC code: " << type << " with " << len << "-byte message.";
    
    // reply with an exception
    send_exception(ioh, IPPC_EX_BAD_IPPC, "Bad IPPC code.");
  };
}

// handle an IPPC close
void event_ippc_close(ippc_handler* ioh) {
  if (ioh == NULL) {
    horde::panic("IPPC server exited unexpectedly.");
  } else if (ioh == session::get_owner()) {
    LOG_EVENT("syscall") << "session owner closed IPPC pipe. closing session.";
    
    session::close(session::get_id(), ioh); // this should eventually call exit
    horde::exit("at 'event_ippc_close'");
  } else {
    LOG_EVENT("syscall") << "client closed.";
    stream::close_all_streams_owned_by(ioh, true);
    remove_statistics_listener(ioh);
  }
}

void event_ippc_owner_flush(ippc_handler* ioh) {
  if (!session::is_connected()) {
    horde::exit("at 'event_ippc_owner_flush'");
  }
}

//
// stream event handlers
//

void notify_syscall_stream_connected(io_principal ioh, string stream_id, int stream_uid) {
  ippcmsg_stream_event evt;
  bzero(&evt, sizeof evt);

  evt.event = evt.CONNECTED;
  strncpy(evt.stream_id, stream_id.c_str(), STREAM_ID_MAX_LENGTH);
  evt.stream_uid = stream_uid;

  ioh->send_message(IPPC_STREAM_EVENT, &evt, sizeof(struct ippcmsg_stream_event));  
}

void notify_syscall_stream_closed(io_principal ioh, string stream_id) {
  ippcmsg_stream_event evt;
  bzero(&evt, sizeof evt);

  evt.event = evt.CLOSED;
  strncpy(evt.stream_id, stream_id.c_str(), STREAM_ID_MAX_LENGTH);
  evt.stream_id[STREAM_ID_MAX_LENGTH] = '\0';
  evt.stream_uid = -1;

  ioh->send_message(IPPC_STREAM_EVENT, &evt, sizeof(struct ippcmsg_stream_event));
}

void notify_syscall_stream_data_ack(io_principal ioh, int stream_uid, 
		      int adu_id, int frag_start, int frag_length, 
		      int rtt_ms) {
  ippcmsg_data_feedback evt;
  bzero(&evt, sizeof evt);

  evt.type = evt.ACK;
  evt.stream_uid = stream_uid;
  evt.adu_id = adu_id;
  evt.frag_start = frag_start;
  evt.frag_length = frag_length;
  evt.rtt_ms = rtt_ms;

  ioh->send_message(IPPC_DATA_FEEDBACK, &evt, sizeof evt);
}

void notify_syscall_stream_data_loss(io_principal ioh, int stream_uid, int adu_id, int frag_start, int frag_length) {
  ippcmsg_data_feedback evt;
  bzero(&evt, sizeof evt);

  evt.type = evt.LOSS;
  evt.stream_uid = stream_uid;
  evt.adu_id = adu_id;
  evt.frag_start = frag_start;
  evt.frag_length = frag_length;
  evt.rtt_ms = 0;

  ioh->send_message(IPPC_DATA_FEEDBACK, &evt, sizeof evt);
}

void notify_syscall_stream_data_recv(io_principal ioh, int stream_uid, int adu_id, int frag_start, int total_length, const char* data, int data_length) {
  size_t header_len = sizeof(struct ippcmsg_data_recv) - 4;
  size_t n = header_len + data_length;
  dynbuf buf(n);
  buf.set_length(n);

  ippcmsg_data_recv* evt = (ippcmsg_data_recv*)buf.data();
  bzero(evt, header_len);

  evt->hstream = stream_uid;
  evt->adu_id = adu_id;
  evt->frag_start = frag_start;
  evt->frag_length = data_length;
  evt->total_length = total_length;
  
  memcpy(evt->data, data, data_length);

  ioh->send_message(IPPC_DATA_RECV, buf);
}

void notify_syscall_stream_choke(io_principal ioh, int stream_uid, bool choke) {
  LOG_EVENT_DEBUG("stream") << "choking " << stream_uid << ":" << choke;
  bool read_enabled = !choke;
  ioh->enable_reading(read_enabled); // stop reading from the pipe until we get an unchoke event...
}

void notify_syscall_stream_flushed(io_principal ioh, int stream_uid) {
  ippcmsg_stream_event evt;
  bzero(&evt, sizeof evt);

  evt.event = evt.FLUSHED;
  evt.stream_id[0] = '\0';
  evt.stream_uid = stream_uid;

  ioh->send_message(IPPC_STREAM_EVENT, &evt, sizeof evt);
}

void notify_syscall_stream_exception(io_principal ioh, string stream_id, stream_exception_code_t code, string ex_msg) {
  // log
  LOG_EVENT("exception") << "stream exception for '" << stream_id << "': '" << ex_msg << "' [" << (int)code << "]";

  // reply with an exception
  ippc_exception_code ixcode;
  switch (code) {
  case EX_STREAM_BAD_ID:
    ixcode = IPPC_EX_BAD_STREAM_ID;
    break;
  case EX_STREAM_INSUFFICIENT_PERMISSION:
    ixcode = IPPC_EX_INSUFFICIENT_PERMISSION;
    break;
  case EX_STREAM_ILLEGAL_OPERATION:
    ixcode = IPPC_EX_ILLEGAL_OPERATION;
    break;
  default:
    ixcode = IPPC_EX_INTERNAL;
  }

  send_exception(ioh, ixcode, ex_msg.c_str());
}

//
// session event handlers
//
void notify_syscall_session_connected(io_principal ioh) {
  ippcmsg_session_event evt;
  bzero(&evt, sizeof evt);

  evt.event = ippcmsg_session_event::CONNECTED;
  strncpy(evt.session_id, session::get_id().c_str(), SESSION_ID_MAX_LENGTH);  

  LOG_EVENT_DEBUG("session") << "session connected.";

  ioh->send_message(IPPC_SESSION_EVENT, &evt, sizeof evt);
}

void notify_syscall_session_closed(io_principal ioh, bool remote) {
  LOG_EVENT_DEBUG("session") << "session closed.";
  
  send_session_closed_ippc(!remote); // notify all my clients

  if (!remote) // exit daemon if this isn't in response to a remote close
    horde::exit("at 'notify_syscall_session_closed'");
}

void notify_syscall_session_exception(io_principal ioh, session_exception_code_t code, string ex_msg) {
  // log
  LOG_EVENT("exception") << "session exception: '" << ex_msg << "' [" << (int)code << "]";

  // reply with an exception
  ippc_exception_code ixcode;
  switch (code) {
  case EX_SESSION_BAD_ID:
    ixcode = IPPC_EX_BAD_STREAM_ID;
    break;
  case EX_SESSION_INSUFFICIENT_PERMISSION:
    ixcode = IPPC_EX_INSUFFICIENT_PERMISSION;
    break;
  case EX_SESSION_ILLEGAL_OPERATION:
    ixcode = IPPC_EX_ILLEGAL_OPERATION;
    break;
  default:
    ixcode = IPPC_EX_INTERNAL;
  }

  send_exception(ioh, ixcode, ex_msg.c_str());
}

void send_session_closed_ippc (bool immediate) {
  // deliver a 'session closed' event to every active ippc handler
  // if (immediate) send right away and then close that handler.
  ippcmsg_session_event evt;
  bzero(&evt, sizeof evt);

  evt.event = ippcmsg_session_event::CLOSED;
  strncpy(evt.session_id, session::get_id().c_str(), SESSION_ID_MAX_LENGTH);

  vector<ippc_handler*> hvec = server->get_children();
  for (vector<ippc_handler*>::iterator iter = hvec.begin(); iter != hvec.end(); iter++) {
    ippc_handler* ioh = *iter;
    if (immediate) {
      ioh->send_message_immediately(IPPC_SESSION_EVENT, &evt, sizeof evt);
      ioh->close(false);
    } else {
      ioh->send_message(IPPC_SESSION_EVENT, &evt, sizeof evt);
    }
  }
}

HORDE_NAMESPACE_END;
