//
// IPPC implementation.
//

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>

#include "ippc.h"
#include "syscall.h"

#include <string>
using namespace std;

//
// debug
//

#ifndef DEBUG_IPPC
#define DEBUG_IPPC 0
#endif

#define DEBUG(...) if (DEBUG_IPPC) { fprintf(stderr, __VA_ARGS__); } else { }

//
// ippc channel context
//

#define MAX_PATH_LENGTH 255
#define CHANNEL_RXTX_BUF_LEN PIPE_BUF
// required: CHANNEL_RXTX_BUF_LEN >= IPPC_MAX_ATOMIC_MESAGE_SIZE + sizeof(ippc_message) - 4

// maximum number of consecutive socket errors.
// (is this the only way to detect closed SEQPACKET connections??)
#define MAX_ERROR_COUNT 10

struct ippc_channel_context {
  char multipart_recv_buffer[IPPC_MAX_MESSAGE_SIZE]; // for reassembling multipart messages

  char rxtx_buffer[CHANNEL_RXTX_BUF_LEN];

  size_t multipart_recv_buf_used;
  
  int seq_last_recv;
  int seq_send;

  int sock_error_count;

  char path[MAX_PATH_LENGTH+1];

  int fd;  
};

struct ippc_message {
  int    seq;
  int    type;
  size_t offset;
  size_t length;
  size_t total_length;

  char   data[1];
};

ippc_channel_context* ippc_channel_context_new(int fd, const char* path) {
  size_t n = sizeof(struct ippc_channel_context);
  ippc_channel_context* newc = (ippc_channel_context*)calloc(1,n);

  DEBUG("context_new: buf at 0x%08x-0x%08x, %d bytes\n", (unsigned int)newc, ((unsigned int)newc)+n-1, n);

  newc->fd = fd;
  newc->seq_last_recv = -1;

  if (path != NULL) {
    strncpy(newc->path, path, MAX_PATH_LENGTH);
  } else {
    newc->path[0] = '\0';
  }
  
  return newc;
}

void ippc_channel_context_delete(ippc_channel_context* c) {
  free(c);
}

int ippc_channel_get_fd(ippc_channel channel) {
  return channel->fd;
}

//
// channel operations (open/accept/close/nonblock)
//

int ippc_channel_is_open(ippc_channel channel) {
  if (channel->sock_error_count > MAX_ERROR_COUNT) return 0;
 
  int e = fcntl(channel->fd, F_GETFL); // a non-mutating fcntl
  if ((e < 0) && (errno == EBADF)) {
    return 0;
  } else {
    return 1; // errno != EBADF or no error
  }
}

ippc_channel_context* ippc_channel_create_from_fd(int fd, const char* path) {
  return ippc_channel_context_new(fd, path);
}

ippc_channel_context* ippc_channel_open(const char* pipe_path) {
  int sock, len;
  struct sockaddr_un remote;

  if (strlen(pipe_path) > MAX_PATH_LENGTH) {
    DEBUG("ERROR: ippc_channel_open: pipe path too long.\n");
    errno = EINVAL;
    return INVALID_CHANNEL;
  }
  
  if ((sock = socket(AF_UNIX, SOCK_SEQPACKET, 0)) == -1) {
    DEBUG("ERROR: ippc_channel_open: socket(...) error %d '%s'\n", errno, strerror(errno));
    return INVALID_CHANNEL;
  }

  memset(&remote, 0, sizeof(remote));
  remote.sun_family = AF_UNIX;
  strcpy(remote.sun_path, pipe_path);

  len = sizeof(remote);
  if (connect(sock, (struct sockaddr *)&remote, len) == -1) {
    DEBUG("ERROR: ippc_channel_open: connect(...) error %d '%s'\n", errno, strerror(errno));
    return INVALID_CHANNEL;
  }

  return ippc_channel_context_new(sock, pipe_path);
}

ippc_channel_context* ippc_channel_accept(const char* pipe_path) {
  int listenfd, ifd;
  socklen_t len;
  struct sockaddr_un remote;
  struct sockaddr_un client;

  if (strlen(pipe_path) > MAX_PATH_LENGTH) {
    DEBUG("ERROR: ippc_channel_accept: pipe path too long.\n");
    errno = EINVAL;
    return INVALID_CHANNEL;
  }
  
  if ((listenfd = socket(AF_UNIX, SOCK_SEQPACKET, 0)) == -1) {
    DEBUG("ERROR: ippc_channel_accept: socket(...) error %d '%s'\n", errno, strerror(errno));
    return INVALID_CHANNEL;
  }

  unlink(pipe_path);
  memset(&remote, 0, sizeof(remote));
  remote.sun_family = AF_UNIX;
  strcpy(remote.sun_path, pipe_path);

  len = sizeof(remote);
  bind(listenfd, (struct sockaddr *)&remote, len);
  
  listen(listenfd, 10);

  if ((ifd = accept(listenfd, (struct sockaddr *)&client, &len)) == -1) {
    DEBUG("ERROR: ippc_channel_accept: accept(...) error %d '%s'\n", errno, strerror(errno));
    close(listenfd);
    return INVALID_CHANNEL;
  }

  close(listenfd);

  return ippc_channel_context_new(ifd, pipe_path);
}

void ippc_channel_close(ippc_channel_context* channel, int remove_from_path) {
  if (channel != INVALID_CHANNEL) {
    // close the pipe
    close(channel->fd);
    
    // unlink the name
    if ((remove_from_path) && (channel->path[0] != '\0')) 
      unlink(channel->path);
    
    // dealloc context
    ippc_channel_context_delete(channel);
  }
}

void ippc_channel_make_nonblock(ippc_channel_context* channel) {
  int flags = fcntl(channel->fd, F_GETFL);
  if (flags >= 0) { // got flags
    flags = flags | O_NONBLOCK;
    fcntl(channel->fd, F_SETFL, flags);
  }
}

//
// send/receive
//
void debug_touch_data(const char* buf, size_t length) {
  int x = 0;

  for (int i = 0; i < length; i++) {
    if (buf[i]) x += 1; // should raise a warning if buf contains uninitialized bytes
  }
}

// Sends buf[offset]...buf[offset+length-1] with the given message
// header fields (seq, offset, total_length).
//
// Returns: -1 on error or atomically sends the given message, and returns length.
// Requires: length < IPPC_MAX_ATOMIC_MESSAGE_SIZE
ssize_t ippc_atomic_send(ippc_channel_context* channel, int seq, int type, const char* buf, 
			 size_t offset, size_t length, size_t total_length) {

  ippc_message* msg = (ippc_message*)channel->rxtx_buffer;

  DEBUG("ippc_atomic_send: seq=%d, type=%d, offset=%d, len=%d\n", seq, type, offset, length);
  
  msg->seq = seq;
  msg->type = type;
  msg->offset = offset;
  msg->length = length;
  msg->total_length = total_length;

  memcpy(msg->data, buf+offset, length);

  size_t header_length = sizeof(ippc_message) - 4; // -4 is for data field...
  ssize_t nsend;
  if ((nsend = send(channel->fd, channel->rxtx_buffer, header_length + length, 0)) < 0) {
    DEBUG("ERROR: ippc_atomic_send: send(...) error %d '%s'\n", errno, strerror(errno));
    channel->sock_error_count++;
  } else {
    nsend -= header_length; // caller doesn't need to know about the header...
    channel->sock_error_count = 0;
  }

  return nsend;
}

ssize_t ippc_message_send(ippc_channel_context* channel, int type, const void* msg, size_t msg_length) {
  // pick a sequence number for this message/message-group
  int seq = channel->seq_send;
  channel->seq_send++;

  DEBUG("ippc_message_send: seq=%d, type=%d, len=%d\n", seq, type, msg_length);

  // send as single or multi-part message...
  if (msg_length <= IPPC_MAX_ATOMIC_MESSAGE_SIZE) {
    // can we do this in an atomic send...
    return ippc_atomic_send(channel, seq, type, (char*)msg, 0, msg_length, msg_length);
  } else {
    // too big, need to use multiple sends...
    size_t offset = 0;
    while (offset < msg_length) {
      size_t n = IPPC_MAX_ATOMIC_MESSAGE_SIZE;
      if ((n + offset) > msg_length) n = msg_length - offset;
      ssize_t x = ippc_atomic_send(channel, seq, type, (char*)msg, offset, n, msg_length);
      if (x < 0) {
	DEBUG("ippc_message_send: multi-part send failed, offset=%d, total=%d.\n", offset, msg_length);
	errno = E2BIG;
	return x;
      } else {
	offset += x;
      }
    }
    return offset;
  }
}

///@todo do we need to put the channel into non-blocking mode for this to work?
char* ippc_message_receive_timed(ippc_channel channel, int* type, size_t* length, int timeout_ms) {
  // assume success
  errno = 0;

  // get a timeval with the right timeout
  timeval sleeptv;
  sleeptv.tv_sec =  timeout_ms / 1000;
  sleeptv.tv_usec = (timeout_ms % 1000) * 1000;

  // set up the file descriptor set
  fd_set readset;
  FD_ZERO(&readset);
  FD_SET(channel->fd, &readset);

  // do select
  select(channel->fd+1, &readset, NULL, NULL, &sleeptv);

  // if the channel file descriptor is readble, try to receive a message...
  if (FD_ISSET(channel->fd, &readset)) 
    return ippc_message_receive(channel, type, length);
  else if ((errno == 0) && (channel->multipart_recv_buf_used > 0))
    return IPPC_RECV_INCOMPLETE;
  else {
    if (errno != 0) DEBUG("I/O: ippc_message_receive_timed error %d: '%s'\n", errno, strerror(errno));
    return NULL;
  }
}

char* ippc_message_receive(ippc_channel channel, int* type, size_t* length) {
  // init
  if (type != NULL)   *type = -1;
  if (length != NULL) *length = 0;
  errno = 0; // assume success

  // atomic receive...
  ssize_t nrecv;
  if ((nrecv = recv(channel->fd, channel->rxtx_buffer, CHANNEL_RXTX_BUF_LEN, 0)) <= 0) {
    channel->sock_error_count++;
    if ((nrecv == 0) && (errno == 0)) {
      errno = EPIPE; // better error number...? ECONNRESET
      return NULL;
    } else if (channel->sock_error_count > MAX_ERROR_COUNT) {
      DEBUG("I/O: socket error count too high in ippc_message_receive, err:'%s' [%d]\n", strerror(errno), errno);
      errno = EBADF; // no longer a useful file descriptor
      return NULL;
    } else if (errno == EAGAIN) {      
      if (channel->multipart_recv_buf_used > 0) {
	DEBUG("I/O: ippc_message_receive has partial multi-part message.\n");
	return IPPC_RECV_INCOMPLETE;
      } else {
	return NULL;
      }
    } else {
      DEBUG("ERROR: ippc_message_receive: recv(...) error %d '%s', nrecv=%d\n", errno, strerror(errno), nrecv);
      return NULL;
    }
  }

  channel->sock_error_count = 0; // reset error count

  // is this part of a multipart message?
  ippc_message* msg = (ippc_message*)channel->rxtx_buffer;
  if (type != NULL)   *type = msg->type;
  if (length != NULL) *length = msg->length;

  DEBUG("ippc_message_receive: recv'd %d bytes, seq=%d, type=%d, len=%d, totlen=%d\n", 
	nrecv, msg->seq, msg->type, msg->length, msg->total_length);

  if (msg->seq != channel->seq_last_recv) { 
    // new message, discard any partially received multi-part message.
    channel->multipart_recv_buf_used = 0;
  }
  channel->seq_last_recv = msg->seq;

  if (msg->length < msg->total_length) { // multipart...
    // copy data to multipart buffer
    char *dest = channel->multipart_recv_buffer + channel->multipart_recv_buf_used;
    memcpy(dest, msg->data, msg->length);
    channel->multipart_recv_buf_used += msg->length;

    if (channel->multipart_recv_buf_used >= msg->total_length) {
      // we're done...
      if (length != NULL) *length = msg->total_length;      
      DEBUG("I/O: ippc_message_receive has complete multi-part message.\n");
      return channel->multipart_recv_buffer;
    } else {
      // need some more recv's, so try another one...
      return ippc_message_receive(channel, type, length);
    }
  } else { 
    return msg->data; 
  }
}

//
// procedure call
//

// won't work with multi-part messages and non-blocking channels.
// requires non-null recv_type and recv_length.
char* ippc_call(ippc_channel channel, 
		int send_type, const void* send_msg, size_t send_msg_length,
		int acceptable_recv_type,
		int *recv_type, size_t* recv_length) {  
  // send
  ssize_t n = ippc_message_send(channel, send_type, send_msg, send_msg_length);
  DEBUG("n=%d, send_msg_length=%d\n", n, send_msg_length);
  if (n != send_msg_length) {
    return NULL;
  }  

  // recv until we get either an exception or a message whose type is
  // equal to acceptable_recv_type
  for (;;) {
    char *buf = ippc_message_receive(channel, recv_type, recv_length);
    if ((buf == NULL) || (buf == IPPC_RECV_INCOMPLETE)) // exit on an error...
      return NULL;

    if ((*recv_type == acceptable_recv_type) || (*recv_type == IPPC_EXCEPTION)) {
      return buf;
    }
  }
}

//
// debugging aid
//

#define HEXBUF_LEN 256
char hexbuf[HEXBUF_LEN+2];

string to_hexstr(const char* data, size_t datalen) {
  if (datalen > HEXBUF_LEN/2) datalen = HEXBUF_LEN/2;

  char *bufptr = hexbuf;
  for (int i = 0; i < datalen; i++) {
    snprintf(bufptr, 3, "%02x", (int)((unsigned char)data[i]));
    bufptr += 2;
    if (((i+1) % 2) == 0) {
      *bufptr = ' ';
      bufptr++;
    }
  }

  *bufptr = '\0';

  return string(hexbuf);
}

string float_to_str(double data, const char* format_str) {
  char *bufptr = hexbuf;
  snprintf(bufptr, HEXBUF_LEN, format_str, data);
  return string(bufptr);
}

string int_to_str(int data, const char* format_str) {
  char *bufptr = hexbuf;
  snprintf(bufptr, HEXBUF_LEN, format_str, data);
  return string(bufptr);
}
