// -*- mode: C++; -*-
/** @file ippc_server.h 
 *
 * An IPPC server implementation. A set of NMSTL I/O handlers that
 * accept and manage open IPPC channels in non-blocking mode. 
 *
 * The Server I/O handler accepts UNIX domain \a SEQPACKET sockets and
 * creates I/O handlers for those sockets.
 *
 * Each socket I/O handler manages a send queue for outgoing messages
 * and calls an external dispatch function to handle incoming
 * messages.
 *
 * This implementation is generic enough to be used by both daemon
 * processes.
 *
 * @note <b>June 2006</b>: refactored, unified \c horde.daemon and \c horde.daemon.session versions
 *
 * @author Asfandyar Qureshi (asfand)
 */

#ifndef __HORDE_IPPC_SERVER_H__
#define __HORDE_IPPC_SERVER_H__

#include <nmstl/io>
#include <nmstl/net>
#include <nmstl/netioevent>

#include <deque>

#include "horde_namespace.h"
#include "ippc.h"
#include "horde_log.h"

using namespace nmstl;
using namespace std;

HORDE_NAMESPACE_BEGIN

class ippc_socket;
class ippc_server;
class ippc_handler;

/**
 * A callback function that handles IPPC messages received on an IPPC channel.
 *
 * For example:
 * \code
 * void my_receive_handler(ippc_handler* event_source, int msg_type, void* msg, size_t msg_length)
 * \endcode
 */
typedef void (*ippc_receive_handler)(ippc_handler*,int,char*,size_t);

/**
 * A callback function that handles IPPC channel close events.
 *
 * Called with a \c NULL argument for server shutdown events.
 *
 * For example:
 * \code
 * void my_close_handler(ippc_handler* event_source)
 * \endcode
 */
typedef void (*ippc_close_handler)(ippc_handler*);

/**
 * A callback function called when the write queue becomes empty.
 */
typedef void (*ippc_flush_handler)(ippc_handler*);

/// an \a ippc_receive_handler that does nothing.
void nop_recv_handler(ippc_handler*,int,char*,size_t);

/// an \a ippc_close_handler that does nothing.
void nop_close_handler(ippc_handler*);

/// an \a ippc_flush_handler that does nothing.
void nop_flush_handler(ippc_handler*);

/**
 * A Unix domain socket of type SOCK_SEQPACKET.
 */
class ippc_socket : public nmstl::socket {
public:
  ippc_socket() : socket() {}
  ippc_socket(address addr, flags f = none) : nmstl::socket(AF_UNIX, SOCK_SEQPACKET, addr, f) {}
  ippc_socket(const iohandle& ioh) : socket(ioh) {}
};

/**
 * I/O handler for events on an open IPPC channel. Manages a send
 * queue for outgoing messages and calls an external dispatch
 * function to handle incoming messages.
 */
class ippc_handler : private io_handler {
private:
  struct send_queue_element {
    int type;
    dynbuf msg;

    send_queue_element(int type, dynbuf msg) : type(type), msg(msg) { }
  };

  deque< send_queue_element* > send_queue;   ///< outgoing message queue.
  ippc_channel ioc;                        ///< underlying IPPC channel.
  ippc_receive_handler recv_handler;       ///< function called to handle receive events.
  ippc_close_handler close_handler;        ///< function called when the socket closes
  ippc_flush_handler flush_handler;        ///< function called when the socket flushes

public:
  /**
   * Creates an IPPC handler.
   *
   * @param ioh The IPPC channel's I/O handle.
   * @param recv_handler Function called to handle message receive events.
   */
  ippc_handler(io_event_loop& loop, iohandle ioh, 
	       ippc_receive_handler recv_handler, ippc_close_handler close_handler,
	       ippc_flush_handler flush_handler) 
    : send_queue(), io_handler(loop, ioh), recv_handler(recv_handler), 
      close_handler(close_handler), flush_handler(flush_handler)
  {  
    ioh.set_blocking(false);
    
    // Create a channel, but don't give it a path name because the server
    // will continue to use that name...
    ioc = ippc_channel_create_from_fd(ioh.get_fd(), NULL);

    // Enable reads and writes.
    want_read(true);
    want_write(true);
  }

  /**
   * Queues a message and tries to send it as soon as possible. Delay
   * is based on when the channel's file descriptor becomes writable.
   *
   * Contents of \a msg should not be modified after this call.
   * 
   * @see ippc_message_send
   */
  void send_message(int type, dynbuf msg) {
    send_queue_element* elem = new send_queue_element(type, msg);
    LOG_EVENT_DEBUG("ippc") << "send_message(" << elem->type << ", msg:" << elem->msg.length() << ") ";    
    send_queue.push_back(elem);
    want_write(true);
  }

  /**
   * Queues and tries to send a message as soon as possible.
   *
   * @note Makes a copy of the given buffer. Use the dynbuf one to
   *       avoid copies.
   */
  void send_message(int type, void* msg, size_t len) {
    dynbuf buf(len);
    memcpy(buf.data(), msg, len);
    buf.set_length(len);

    this->send_message(type, buf);
  }

  
  /**
   * Sends the message immediately, before returning. May block until I/O is
   * completed.
   */
  void send_message_immediately(int type, void* msg, size_t len) {
    get_ioh().set_blocking(true);
    ssize_t n = ippc_message_send(ioc, type, msg, len);
    get_ioh().set_blocking(false);
  }

  /**
   * Shuts down this I/O handler and calls the close event handler.
   */
  void close(bool call_close_handler=true);

  /**
   * Get the file descriptor for the underlying pipe.
   *
   * @warning Be very careful in how you use this descriptor.
   */
  int get_fd() {
    return get_ioh().get_fd();
  }

  /**
   * Stops this handler from reading messages; or tells it to start 
   * reading again. By default reading is enabled when a handler is
   * created.
   */
  void enable_reading(bool enabled) {
    want_read(enabled);
  }
   
protected:
  void ravail();
  void wavail();
};

/**
 * IPPC client acceptor. Unix pipe I/O acceptor. Accepts SEQPACKET
 * sockets and creates \a ippc_handler objects for each new socket.
 *
 * @todo this code never deletes defunct ippc_handler objects
 */
class ippc_server : public io_handler{
private:
  string path;
  ippc_socket sock;
  ippc_receive_handler recv_handler;
  ippc_close_handler close_handler;
  ippc_flush_handler flush_handler;

  vector<ippc_handler*> handlers;
public:
  /**
   * Create an IPPC acceptor bound to a specific path.
   *
   * @param pat A complete path. The entire path, except for the terminal filename,
   *            must name a valid location in the filename namespace.
   *
   * @param recv_handler
   *            The function that will be called when IPPC messages are received on
   *            any of the \a ippc_handler objects spawned by this server.
   *
   * @param close_handler
   *            The function that will be called when any of the \a ippc_handler 
   *            objects spawned by this server close.
   */
  ippc_server(io_event_loop& loop, string path, 
	      ippc_receive_handler recv_handler=nop_recv_handler, 
	      ippc_close_handler close_handler=nop_close_handler,
	      ippc_flush_handler flush_handler=nop_flush_handler);

  /**
   * Cleanly shut down the server.
   */
  void stop_server(bool call_close_handler=true);

  /**
   * Creates an \a ippc_handler tied to this server, using an appropriate open 
   * socket file descriptor.
   *
   * @param fd Open file descriptor for an \a AF_UNIX / \a SOCK_SEQPACKET socket. 
   *           The socket should not be used after calling this function.
   *
   * @param recv_handler 
   *           Function called when IPPC messages are received on the given socket.
   *           If this is \a NULL, this server's default \a recv_handler is used.
   *
   * @param close_handler 
   *           Function called when the socket closes.
   *           If this is \a NULL, this server's default \a close_handler is used.
   *
   * @return The newly created \a ippc_handler object.
   */
  ippc_handler* create_ippc_handler(int fd, 
				    ippc_receive_handler recv_handler=NULL, 
				    ippc_close_handler close_handler=NULL,
				    ippc_flush_handler flush_handler=NULL);

  
  /** Get the path this server is bound to. */
  string get_path() { return this->path; }

  /// returns the handlers created by this server.
  vector<ippc_handler*> get_children() {
    return handlers; // return a copy of the children vector...
  }

protected:
  virtual void ravail();
};

HORDE_NAMESPACE_END

#endif
