#!/usr/bin/python
#
# Simple make manager. Functions for building binaries.
#
# Usage:
#    make.py <binaryname> <src1> <src2> <src3> ...
# Or:
#    make.py clean <src1> <src2> <src3> ...
#
import sys
import os

LINKER = "g++"
LINKER_FLAGS = "-lstdc++ -lnmstl -lreadline -lexpat -lncurses -lpthread"
COMPILER_WRAPPER="wtf"
COMPILER = "g++"
COMPILER_FLAGS = "-g -c"
OBJ_FILE_DIR = "obj"

def split_filename(filename):
    """
    returns (name, extension)
    """
    name=""
    ext=""
    try:
        period = filename.rindex('.')
        name = filename[:period]
        ext = filename[period+1:]
    except:
        pass

    return name,ext

def get_header_files(filename, src_ext):
    """
    returns a list of header files.
    """
    try:
        srcfile = open(filename + "." + src_ext, 'r')

        n = 0
        hfiles = []
        for line in srcfile:
            line = line.strip()
            
            if line[:10] == '#include "':
                    x = line.split('"')
                    hfiles.append(x[1])
                    
            n += 1
            if n > 64: break

        srcfile.close()
        
        return hfiles
    except:
        return []

def get_objfilename(name):
    # assumes no extension for given filename
    try:
        name = name[name.rindex("/")+1:]
    except:
        pass
    
    return OBJ_FILE_DIR + '/' + name + '.o'

def build_succeeded(filename, src_ext,src_lastmod):
      try:
          obj_lastmod = os.stat(get_objfilename(filename))[8]
          #src_lastmod = os.stat(filename + "." + src_ext)[8]
          return (src_lastmod <= obj_lastmod)        
      except:
          return False

def get_header_lastmod(h, inc_paths):
    try:
        h_lastmod = os.stat(h)[8]  # no extra path
    except:
        h_lastmod = 0

    for i in inc_paths:
        try:
            lastmod = os.stat(i + "/" + h)[8]
            if lastmod > h_lastmod: h_lastmod = lastmod
        except:
            pass
        

    return h_lastmod;
          
def get_updated_dependency(filename, src_ext, inc_paths):
    """
    Assumes source is either c or cc
    
    Returns a message specifying a filename
    (a changed dependency) if there is one, or
    returns "". Returns 'clean compile' when no .o
    file exists.
    
    Returns non-empty when:
    - no object file
    - src file is newer than the object file
    - src file a header file has changed.    
    """
    try:
        # Get last mod time for object file
        try:
            obj_lastmod = os.stat(get_objfilename(filename))[8]
        except: 
            return 'clean compile',None

        # Is obj file newer than source, return right away
        # otherwise test header files.
        src_lastmod = os.stat(filename + '.' + src_ext)[8]

        if src_lastmod > obj_lastmod:
            return "updated: " + filename + '.' + src_ext, src_lastmod
        else:
           # Test dependencies
           header_files = get_header_files(filename, src_ext)

           for h in header_files:
               h_lastmod = get_header_lastmod(h, inc_paths);
               
               if h_lastmod > obj_lastmod:
                   return "updated: " + h, h_lastmod
                
           # if we reach this point, nothing new...
           return "",None
        
    except:
        return "python error",None
    

def make_clean(src_files):
    """
    Deletes any .o file that matches an element of <src_files>.

    Also removes OBJ_FILE_DIR
    """
    print "Deleting object files..."
    for src_filename in src_files:
        try:
            name,ext = split_filename(src_filename)
            obj_filename = get_objfilename(name)

            f = open(obj_filename, 'rb') # throws an exception if file-not-found
            f.close()

            os.remove(obj_filename)
            print "+ del", obj_filename
            
        except:    
            pass

    try:
        os.rmdir(OBJ_FILE_DIR)
    except:
        pass
    
    return True
    print "Done."

def init_obj_file_dir():
    try:
        os.mkdir(OBJ_FILE_DIR)
    except:
        pass

def make_binary(target, src_files, inc_paths=[]):
    init_obj_file_dir();
    print "\033[1;1m>>> Starting compile process for '%s' <<<\033[0m" % target
    
    # init needed vars
    link_command = LINKER.split() + LINKER_FLAGS.split()
    compile_command = COMPILER_WRAPPER.split() + COMPILER.split() + COMPILER_FLAGS.split() + ["-I" + x for x in inc_paths]
    
    # build each source file
    obj_files = []
    for src in src_files:
        fname,ext = split_filename(src)
        msg,lastmod = get_updated_dependency(fname,ext, inc_paths)
        out_file = get_objfilename(fname)
        if msg: # need to recompile
            print "+ CC " + src + " (" + msg + ")"
            
            os.spawnvp(os.P_WAIT, compile_command[0], compile_command + [src, '-o', out_file])

            if not build_succeeded(fname,ext,lastmod):
                print "\033[1;1m>>> compile failed for '" + src + "', exiting <<<\033[0m"
                print ">>> make sure all the compiler commands including 'wtf' are in your path <<<"
                return False
        else:
            print "  == " + src
            
        obj_files.append(out_file)
            
    # link
    print "+ LD " + target

    command = link_command + obj_files + ["-o", target]
    os.spawnvp(os.P_WAIT, command[0], command)

    # done?
    try:
        f = open(target, 'rb') # throws an exception if file-not-found
        f.close()

        print "\033[1;1mCREATED", target, "\033[0m"
        print "Done."
        return True        
    except:
        print ">>> linking failed. exiting. <<<"
        return False

def add_macro_definitions(macro_list):
    global COMPILER_FLAGS # ouch, no longer a constant...
    for m in macro_list:
        COMPILER_FLAGS += " -D " + str(m)

def make_main(src=[]):
    if len(sys.argv) < 2:
        print "Usage:", sys.argv[0], "<target | clean | rebuild> <src1> [src2] [src3] ..."
    else:
        target = sys.argv[1]
        src_files = src + sys.argv[2:]
        
        if target == "clean":
            return make_clean(src_files)
        elif target == "x" or target == "rebuild":
            make_clean(src_files)
            return make_binary(target, src_files)
        else:
            return make_binary(target, src_files)


#if __name__ == "__main__":
try:
    suppress_make_main
except NameError:
    make_main()
else:
    pass # told to suppress main function...
