// -*- mode: C++; -*-
/** @file horde_log.h 
 *
 * Horde event logging library. Provides functionality to log 
 * timestamped messages. Particularly useful for debugging purposes.
 * Any logging statements can be disabled at compile time by using an
 * appropriate \c \#define. Disabled logging statements have zero, or
 * very low overhead.
 *
 * Three logging macros are defined: \a LOG_NET, \a LOG_EVENT, and \a
 * LOG_EVENT_DEBUG. \a LOG_NET writes messages to the network log, and
 * the other two write to the 'event' log.
 *
 * A message may be logged as follows:
 *
 * \code
 * LOG_EVENT("ippc") << "socket write failed: " << strerror(errno) << ", errno=" << errno;
 * \endcode
 *
 * Here \c "ippc" specifies the event category. A category should be a single word, 
 * not more than 10 characters long.
 *
 * You can write values to a \a LOG_EVENT in the same way that you can write values to
 * \a cout, but \a LOG_EVENT is not an \a ostream object.
 * 
 * \b Initialization:
 * You \b must call logging_init() before using any of the \a LOG_ macros.
 * 
 * \b Disabling: 
 * You can \c \#define one or more of the following symbols (before including this file) 
 * to modify the behaviour of logging statements:
 * 
 * \arg \a HORDE_LOG_DISABLE_ALL:         Disable all logging statements.
 * \arg \a HORDE_LOG_DISABLE_EVENT:       Disable \a LOG_EVENT and \a LOG_EVENT_DEBUG statements.
 * \arg \a HORDE_LOG_DISABLE_NET:         Disable \a LOG_NET statements.
 * \arg \a HORDE_LOG_ENABLE_EVENT_DEBUG:  Enable \a LOG_EVENT_DEBUG statements.
 * \arg \a HORDE_LOG_DISABLE_CODE_INFO:   Avoid logging file and line-number information.
 *
 * @author Asfandyar Qureshi (asfand)
 * 
 * @note Some parts of this are based on the NMSTL debug code.
 * @todo dealing with log flushing (set up a timer to periodically flush?)
 */

#ifndef __HORDE_LOG_H__
#define __HORDE_LOG_H__

#include <fstream>
#include <string>
#include <nmstl/ntime>

#include "horde_namespace.h"

using namespace std;
using namespace nmstl;

HORDE_NAMESPACE_BEGIN

//
// logger classes
//

/**
 * A log message.
 */
class log_msg {
private:
  const char *file;
  int line;
  bool code_info;
  const char* event_id;

  string str;

public:

  /**
   * An initially empty message with an event category id, and
   * information on where in the code this \a log_msg object was
   * created (file and line number).
   *
   * @param file Source file in where this constructor was called.
   * @parma line Line number where this was called.
   * @param code_info Whether or not to write the \a file and \a line as part of the message.
   * @param event_id An event category, not more than 10 characters long.
   */
  log_msg(const char* file, int line, bool code_info, const char* event_id)
    : file(file), line(line), code_info(code_info), event_id(event_id)
  {}


  /// Convert the given value to a string and append it to the message.
  template<class T>
  log_msg& operator << (const T& t) { str += to_string(t); return *this; }

  /// Append a string to the message.
  log_msg& operator << (const string& s) { str += s; return *this; }

  /// Append a string to the message.
  log_msg& operator << (const char* s) { 
    if (s == NULL)
      str += "<null>";
    else
      str += s; 

    return *this; 
  }

  /// Append a boolean to the message.
  log_msg& operator << (bool b) { 
    if (b) str += "true"; 
    else str += "false"; 

    return *this; 
  }

  /// Append a pointer to the message
  log_msg& operator << (void* p) { 
    char buf[16];
    snprintf(buf, 16, "ptr:%08x", (int)p);
    str += buf;

    return *this;
  }

  /// Write the complete message line to the given \a ostream.
  void write(ostream& o) const;
};

class horde_net_packet;

/**
 * A NET log message.
 */
class net_log_msg {
private:
  const char *file;
  int line;
  bool code_info;
  const char* event_id;

  string str;
public:

  /**
   * An initially empty net log message with an event category id, and
   * information on where in the code this \a log_msg object was
   * created (file and line number).
   *
   * @param file Source file in where this constructor was called.
   * @parma line Line number where this was called.
   * @param code_info Whether or not to write the \a file and \a line as part of the message.
   * @param event_id An event category, not more than 10 characters long.
   */
  net_log_msg(const char* file, int line, bool code_info, const char* event_id)
    : file(file), line(line), code_info(code_info), event_id(event_id)
  {}

  /// Append a string to the message.
  net_log_msg& operator << (horde_net_packet& packet);

  /// Convert the given value to a string and append it to the message.
  template<class T>
  net_log_msg& operator << (const T& t) { str += to_string(t); return *this; }

  /// Append a string to the message.
  net_log_msg& operator << (const string& s) { str += s; return *this; }

  /// Append a boolean to the message.
  net_log_msg& operator << (bool b) { 
    if (b) str += "true"; 
    else str += "false"; 

    return *this; 
  }

  /// Write the complete message line to the given \a ostream.
  void write(ostream& o) const;
};

/**
 * A log message writer. Writes log messages to a file.
 */
class logger {
  ofstream file;

public: 
  /// Create a logger, in a closed state.
  logger() { }

  /// Open a file to write messages to.
  void open(const char* filename) { 
    file.open (filename, ofstream::out | ofstream::app);
  }

  /// Test if logger is open, or closed.
  bool is_open() { return file.is_open(); }

  //// Close the logger if it is open.
  void close() {
    file.close();
  }

  //// Flush buffered writes.
  void flush() {
    file.flush();
  }    

  /**
   * Write a log message. Operator overloading trick to make
   * disabled logging statements efficient. See the definition of \a
   * LOG_NET to understand this hack. Borrowed from NMSTL.
   */
  bool operator ^ (const log_msg& msg);

  bool operator ^ (const net_log_msg& msg);
};

// actual logs
extern logger horde_net_logger;       ///< log writer for LOG_NET
extern logger horde_event_logger;     ///< log writer for LOG_EVENT and LOG_EVENT_DEBUG

/**
 * Initialize logging. Creates two new log files, one for events and
 * one for network messages. Always creates new log files, by using
 * the current time as a filename suffix.
 * 
 * @param base_path The log files will be created in the \c
 * logs/event and \c logs/net subdirectories of \a base_path.
 * These directories must exist.
 *
 * @param prefix Log filenames start with this string.
 *
 * @return \a true if log files were created and writers were
 * initialized successfully.
 */
bool logging_init(const char *prefix, const char *base_path);

/** 
 * End logging. Close log writers. \a LOG_ macros will not work
 * after this is called.
 */
void logging_close();

/// Flush buffered writes.
void logging_flush();

/** Log a critical error message and exit the program. */
void panic(const char* msg="");

/** Log a final message and exit the program. */
void exit(const char* msg="");

#ifndef DOXYGEN_SKIP
//
// macro definitions to control what is logged.
//

// use instead of false to avoid "statement with no effect" warning
inline bool __horde_false() { return false; }

#ifndef HORDE_LOG_DISABLE_ALL

#  ifdef HORDE_LOG_ENABLE_FLUSH
#    define LOG_FLUSH true
#  else
#    define LOG_FLUSH __horde_false()
#  endif

#  ifdef HORDE_LOG_DISABLE_NET
#    define LOG_NET_ENABLED __horde_false()
#  else
#    define LOG_NET_ENABLED true
#  endif

#  ifdef HORDE_LOG_DISABLE_EVENT
#    define LOG_EVENT_ENABLED __horde_false()
#    define LOG_EVENT_DEBUG_ENABLED __horde_false()
#  else
#    define LOG_EVENT_ENABLED true

#    ifdef HORDE_LOG_ENABLE_EVENT_DEBUG
#      define LOG_EVENT_DEBUG_ENABLED true
#    else
#      define LOG_EVENT_DEBUG_ENABLED __horde_false()
#    endif

#  endif

#  ifdef HORDE_LOG_DISABLE_CODE_INFO
#    define CODE_INFO false
#  else
#    define CODE_INFO true
#  endif

// macros used for actual output

#  define LOG_NET(event_type) (LOG_NET_ENABLED) && horde_net_logger ^ net_log_msg(__FILE__, __LINE__, false, event_type)
#  define LOG_EVENT(event_type) (LOG_EVENT_ENABLED) && horde_event_logger ^ log_msg(__FILE__, __LINE__, CODE_INFO, event_type)
#  define LOG_EVENT_DEBUG(event_type) (LOG_EVENT_DEBUG_ENABLED) && horde_event_logger ^ log_msg(__FILE__, __LINE__, CODE_INFO, event_type)

#else

// no logging

#  define LOG_NET(event_type) __horde_false() && horde_net_logger ^ net_log_msg(NULL, 0, false, NULL)
#  define LOG_EVENT(event_type) __horde_false() && horde_event_logger ^ log_msg(NULL, 0, false, NULL)
#  define LOG_EVENT_DEBUG(event_type) __horde_false() && horde_event_logger ^ log_msg(NULL, 0, false, NULL)

#endif

#endif

HORDE_NAMESPACE_END

#endif
