/** @file objectives.h 
 *
 * ADU objectives. Structures shared between daemon and API.
 */

#ifndef __HORDE_OBJECTIVES_H__
#define __HORDE_OBJECTIVES_H__

struct horde_objective_tags {
 
  /**
   * Relative priority of data. Lower numbers are higher
   * priorities. Higher priority data is sent in preference to lower
   * priority data.
   */
  int priority;

  /**
   * Whether or not this data can be fragmented during packetization
   * (boolean: 0 or non-zero). If this is zero, the data will never be
   * fragmented.
   */
  int fragmentation_allowed;

  /** 
   * The correlation group of this data. Ignored if equal to zero. If
   * non-zero, Horde will try to schedule two ADU's with the same
   * correlation_group in packets that are not likely to experience
   * correlated losses.
   */ 
  int correlation_group;

  /**
   * If this objective is non-zero, the data may be transmitted
   * multiple times, if the available bandwidth is greater than the
   * offered load.
   *
   * @note only valid when fragmentation_allowed is 0
   */
  int multiple_tx;

  /**
   * Maximum allowed loss probability (as a percentage 0-100).  This
   * data will not be sent in slots with a higher loss probablity. A
   * low value may result in head-of-line blocking.
   */
  int loss_threshold;

  /**
   * If a slot's expected RTT is greater than this, it will not be
   * used.  If this is zero, it is ignored. A low value may result in
   * head-of-line blocking.
   */
  int rtt_threshold;  

  /**
   * Maximum allowed loss probability, expressed as the difference from
   * the mean loss probability over all channels (as a percentage 0-100).
   * If this is zero, it is ignored.
   */
  int relative_loss_threshold;

  /**
   * A TX slot whose expected RTT is greater than the average RTT of
   * the stream plus this threshold will not be used. If this is zero,
   * it is ignored.
   */
  int relative_rtt_threshold;

  horde_objective_tags(int priority, int frag_allowed, int corr_group, int multiple_tx, 
		       int loss_thresh, int rtt_threshold, int rel_loss_threshold, int rel_rtt_threshold) :
    priority(priority), fragmentation_allowed(frag_allowed), correlation_group(corr_group), multiple_tx(multiple_tx),
    loss_threshold(loss_thresh), rtt_threshold(rtt_threshold), relative_loss_threshold(rel_loss_threshold),
    relative_rtt_threshold(rel_rtt_threshold)
  { }

  /** 
   * Inits tags to default values.
   *
   * \code
   * priority = 0
   * fragmentation_allowed = 1
   * correlation_group = 0 
   * multiple_tx = 0
   * loss_threshold = 100
   * rtt_threshold = 0
   * relative_loss_threshold = 0
   * relative_rtt_threshold = 0
   * \endcode
   */
  horde_objective_tags() :
    priority(0), fragmentation_allowed(1), correlation_group(0), multiple_tx(0),
    loss_threshold(100), rtt_threshold(0), relative_loss_threshold(0), relative_rtt_threshold(0)
  { }
};

#endif
