#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <iostream>
#include <nmstl/debug>

#include "util.h"

using namespace std;
using namespace nmstl;

int max_retry = 0;
int retry_reset_period = 600;
int retry_sleep = 10;
char *iface = NULL;
bool help_requested = false;
bool reset_all = false;

int current_child_pid = 0;

argument_table_element argument_table[] = {
  { ARG_TYPE_BOOLEAN, "help", &help_requested, false, false, "Print out this help." },
  { ARG_TYPE_STRING, "iface", &iface, true, false, 
    "The interface to manage. This is the internal interface name. "
    "The command 'scripts/iface_X_up' will be used to try to connect this interface, " 
    "where X is the value of this argument."},
  { ARG_TYPE_INTEGER, "max-retry", &max_retry, false, false, 
    "Maximum number of reconnection attempts during a long period of time. "
    "If 0, then an infinite number of retries are assumed. Default is 0." },
  { ARG_TYPE_INTEGER, "retry-reset-period", &retry_reset_period, false, false,
    "Period in seconds used by the max-retry logic. After this many seconds, the "
    "reconnection counter is reset to zero. Default is 600."},
  { ARG_TYPE_INTEGER, "retry-sleep", &retry_sleep, false, false,
    "Delay in seconds between a disconnection and the attempt to reconnect. "
    "Defaults to 10."},
  { ARG_TYPE_BOOLEAN, "reset-all", &reset_all, false, false, 
    "Whether or not 'scripts/reset_all' should be called before each reconnection attempt. "},
  END_OF_TABLE
};

static void handle_signal(int s) {
  if (s == SIGUSR1) {
    if (current_child_pid > 0) { // if a current child exists, kill it
      kill(current_child_pid, SIGKILL);
    }
  } else {
    INFO << "received a termination signal. exiting.";
    exit(0);
  }
}

main (int argc, char** argv) {
  bool ok = arguments_parse(argv, argc, argument_table);

  if (help_requested) {
      arguments_print_help(argv[0], argument_table);
  } else if (!ok) {
    arguments_print_usage(argv[0], argument_table);
  } else {
    // info
    INFO << "managing interface '" << iface << "', max_retry=" << max_retry 
	 << ", retry_sleep=" << retry_sleep << "s, retry_reset_period=" << retry_reset_period << "s";      

    // derive name of interface manager script     
    char *cwd = get_current_dir_name();
    string iface_script = cwd;    

    iface_script += "/scripts/iface_";
    iface_script += iface;
    iface_script += "_up";

    string reset_all_script = cwd;
    reset_all_script += "/scripts/reset_all";

    free(cwd);

    DEBUG << "using up script: '" << iface_script << "' and reset script: '" << reset_all_script << "'";

    // register signal handlers
    signal(SIGUSR1, handle_signal);    
    signal(SIGKILL, handle_signal);    
    siginterrupt(SIGUSR1, 1);

    // enter interface manager loop
    int retry_count = 0;
    int attempt = 0;
    time_t last_period = time(NULL);
    timespec sleep_period = { retry_sleep, 0 };
    for (;;) {
      attempt++;
      tty_bold_on();printf("Attempting to connect...");tty_bold_off();
      printf("\nattempt #%d [%d attempts this period]\n", attempt, retry_count+1);

      // call reset_all ?
      if (reset_all) {
	tty_bold_on();printf("calling reset_all...");tty_bold_off();
	printf("\n");
	if (system(reset_all_script.c_str()) < 0) {
	  CRIT << "could not run reset_all script. continuing anyway.";
	}
      }

      // try to bring the interface up...
      tty_bold_on();printf("calling interface up script...");tty_bold_off();
      printf("\n");
      current_child_pid = spawn(iface_script);
      if (current_child_pid > 0) {
	int status=0;
	pid_t x = wait(&status);
      } else {
	CRIT << "could not start interface script.";
      }

      tty_bold_on();printf("Disconnected.");tty_bold_off();
      printf("\n");

      // post-disconnection logic
      time_t now = time(NULL);                      
      if (now - last_period > retry_reset_period) {  // reset retry counter?
	retry_count = 0;
	last_period = now;
      }

      retry_count++;
      if ((max_retry) && (retry_count >= max_retry)) break; // too many retries?

      printf("sleeping for %d seconds.\n", retry_sleep);
      nanosleep(&sleep_period, NULL);                // inter-reconnect sleep      

    }
  }
}
